import concurrent.futures

from .._sleipnir.optimization import *


def multistart(solve, initial_guesses):
    """
    Solves an optimization problem from different starting points in parallel,
    then returns the solution with the lowest cost.

    Each solve is performed on a separate thread. Solutions from successful
    solves are always preferred over solutions from unsuccessful solves, and
    cost (lower is better) is the tiebreaker between successful solves.

    Parameter ``solve``:
        A user-provided function that takes a decision variable initial guess
        and returns a MultistartResult.

    Parameter ``initial_guesses``:
        A list of decision variable initial guesses to try.
    """
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=len(initial_guesses)
    ) as executor:
        futures = [
            executor.submit(solve, initial_guess) for initial_guess in initial_guesses
        ]
        results = [
            future.result() for future in concurrent.futures.as_completed(futures)
        ]

    # Prioritize successful solve, otherwise prioritize solution with lower cost
    return min(
        results,
        key=lambda x: (
            int(x[0] != ExitStatus.SUCCESS),
            x[1],
        ),
    )



import numpy as np
import pytest
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_double_integrator_problem():
    TOTAL_TIME = 3.5
    dt = 0.005
    N = int(TOTAL_TIME / dt)

    r = 2.0

    problem = Problem()

    # 2x1 state vector with N + 1 timesteps (includes last state)
    X = problem.decision_variable(2, N + 1)

    # 1x1 input vector with N timesteps (input at last state doesn't matter)
    U = problem.decision_variable(1, N)

    # Kinematics constraint assuming constant acceleration between timesteps
    for k in range(N):
        t = dt
        p_k1 = X[0, k + 1]
        v_k1 = X[1, k + 1]
        p_k = X[0, k]
        v_k = X[1, k]
        a_k = U[0, k]

        # pₖ₊₁ = pₖ + vₖt + 1/2aₖt²
        problem.subject_to(p_k1 == p_k + v_k * t + 0.5 * a_k * t**2)

        # vₖ₊₁ = vₖ + aₖt
        problem.subject_to(v_k1 == v_k + a_k * t)

    # Start and end at rest
    problem.subject_to(X[:, :1] == np.array([[0.0], [0.0]]))
    problem.subject_to(X[:, N : N + 1] == np.array([[r], [0.0]]))

    # Limit velocity
    problem.subject_to(-1 <= X[1:2, :])
    problem.subject_to(X[1:2, :] <= 1)

    # Limit acceleration
    problem.subject_to(-1 <= U)
    problem.subject_to(U <= 1)

    # Cost function - minimize position error
    problem.minimize(sum((r - X[0, k]) ** 2 for k in range(N + 1)))

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    A = np.array([[1.0, dt], [0.0, 1.0]])
    B = np.array([[0.5 * dt * dt], [dt]])

    # Verify initial state
    assert X.value(0, 0) == pytest.approx(0.0, abs=1e-8)
    assert X.value(1, 0) == pytest.approx(0.0, abs=1e-8)

    # Verify solution
    x = np.zeros((2, 1))
    u = np.zeros((1, 1))
    for k in range(N):
        # Verify state
        assert X.value(0, k) == pytest.approx(x[0, 0], abs=1e-2)
        assert X.value(1, k) == pytest.approx(x[1, 0], abs=1e-2)

        # Determine expected input for this timestep
        if k * dt < 1.0:
            # Accelerate
            u[0, 0] = 1.0
        elif k * dt < 2.05:
            # Maintain speed
            u[0, 0] = 0.0
        elif k * dt < 3.275:
            # Decelerate
            u[0, 0] = -1.0
        else:
            # Accelerate
            u[0, 0] = 1.0

        # Verify input
        if (
            k > 0
            and k < N - 1
            and abs(U.value(0, k - 1) - U.value(0, k + 1)) >= 1.0 - 1e-2
        ):
            # If control input is transitioning between -1, 0, or 1, ensure it's
            # within (-1, 1)
            assert U.value(0, k) >= -1.0
            assert U.value(0, k) <= 1.0
        else:
            assert U.value(0, k) == pytest.approx(u[0, 0], abs=1e-4)

        # Project state forward
        x = A @ x + B @ u

    # Verify final state
    assert X.value(0, N) == pytest.approx(r, abs=1e-8)
    assert X.value(1, N) == pytest.approx(0.0, abs=1e-8)

    # Log states for offline viewing
    with open("Double integrator states.csv", "w") as f:
        f.write("Time (s),Position (m),Velocity (rad/s)\n")

        for k in range(N + 1):
            f.write(f"{k * dt},{X.value(0, k)},{X.value(1, k)}\n")

    # Log inputs for offline viewing
    with open("Double integrator inputs.csv", "w") as f:
        f.write("Time (s),Acceleration (m/s²)\n")

        for k in range(N + 1):
            if k < N:
                f.write(f"{k * dt},{U.value(0, k)}\n")
            else:
                f.write(f"{k * dt},0.0\n")



import numpy as np
import pytest
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_unconstrained1d():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(2.0)

    problem.minimize(x * x - 6.0 * x)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(3.0, abs=1e-6)


def test_unconstrained2d_1():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(1.0)
    y = problem.decision_variable()
    y.set_value(2.0)

    problem.minimize(x * x + y * y)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(0.0, abs=1e-6)
    assert y.value() == pytest.approx(0.0, abs=1e-6)


def test_unconstrained2d_2():
    problem = Problem()

    x = problem.decision_variable(2)
    x[0].set_value(1.0)
    x[1].set_value(2.0)

    problem.minimize(x.T @ x)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value(0) == pytest.approx(0.0, abs=1e-6)
    assert x.value(1) == pytest.approx(0.0, abs=1e-6)


# Maximize xy subject to x + 3y = 36.
#
# Maximize f(x,y) = xy
# subject to g(x,y) = x + 3y - 36 = 0
#
#         value func  constraint
#              |          |
#              v          v
# L(x,y,λ) = f(x,y) - λg(x,y)
# L(x,y,λ) = xy - λ(x + 3y - 36)
# L(x,y,λ) = xy - xλ - 3yλ + 36λ
#
# ∇_x,y,λ L(x,y,λ) = 0
#
# ∂L/∂x = y - λ
# ∂L/∂y = x - 3λ
# ∂L/∂λ = -x - 3y + 36
#
#  0x + 1y - 1λ = 0
#  1x + 0y - 3λ = 0
# -1x - 3y + 0λ + 36 = 0
#
# [ 0  1 -1][x]   [  0]
# [ 1  0 -3][y] = [  0]
# [-1 -3  0][λ]   [-36]
#
# Solve with:
# ```python
#   np.linalg.solve(
#     np.array([[0,1,-1],
#               [1,0,-3],
#               [-1,-3,0]]),
#     np.array([[0], [0], [-36]]))
# ```
#
# [x]   [18]
# [y] = [ 6]
# [λ]   [ 6]
def test_equality_constrained_1():
    problem = Problem()

    x = problem.decision_variable()
    y = problem.decision_variable()

    problem.maximize(x * y)

    problem.subject_to(x + 3 * y == 36)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(18.0, abs=1e-5)
    assert y.value() == pytest.approx(6.0, abs=1e-5)


def test_equality_constrained_2():
    problem = Problem()

    x = problem.decision_variable(2)
    x[0].set_value(1.0)
    x[1].set_value(2.0)

    problem.minimize(x.T @ x)

    problem.subject_to(x == np.array([[3.0], [3.0]]))

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value(0) == pytest.approx(3.0, abs=1e-5)
    assert x.value(1) == pytest.approx(3.0, abs=1e-5)


def test_inequality_constrained_2d():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(5.0)
    y = problem.decision_variable()
    y.set_value(5.0)

    problem.minimize(x * x + y * 2 * y)
    problem.subject_to(y >= -x + 5)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(3.0 + 1.0 / 3.0, abs=1e-6)
    assert y.value() == pytest.approx(1.0 + 2.0 / 3.0, abs=1e-6)



"""This problem tests the case where regularization fails"""

import math

import numpy as np
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_arm_on_elevator_problem():
    N = 800

    ELEVATOR_START_HEIGHT = 1.0  # m
    ELEVATOR_END_HEIGHT = 1.25  # m
    ELEVATOR_MAX_VELOCITY = 1  # m/s
    ELEVATOR_MAX_ACCELERATION = 2.0  # m/s²

    ARM_LENGTH = 1.0  # m
    ARM_START_ANGLE = 0.0  # rad
    ARM_END_ANGLE = math.pi  # rad
    ARM_MAX_VELOCITY = 2.0 * math.pi  # rad/s
    ARM_MAX_ACCELERATION = 4.0 * math.pi  # rad/s²

    END_EFFECTOR_MAX_HEIGHT = 1.8  # m

    TOTAL_TIME = 4.0  # s
    dt = TOTAL_TIME / N

    problem = Problem()

    elevator = problem.decision_variable(2, N + 1)
    elevator_accel = problem.decision_variable(1, N)

    arm = problem.decision_variable(2, N + 1)
    arm_accel = problem.decision_variable(1, N)

    for k in range(N):
        # Elevator dynamics constraints
        problem.subject_to(
            elevator[0, k + 1]
            == elevator[0, k] + elevator[1, k] * dt + 0.5 * elevator_accel[0, k] * dt**2
        )
        problem.subject_to(
            elevator[1, k + 1] == elevator[1, k] + elevator_accel[0, k] * dt
        )

        # Arm dynamics constraints
        problem.subject_to(
            arm[0, k + 1] == arm[0, k] + arm[1, k] * dt + 0.5 * arm_accel[0, k] * dt**2
        )
        problem.subject_to(arm[1, k + 1] == arm[1, k] + arm_accel[0, k] * dt)

    # Elevator start and end conditions
    problem.subject_to(elevator[:, :1] == np.array([[ELEVATOR_START_HEIGHT], [0.0]]))
    problem.subject_to(
        elevator[:, N : N + 1] == np.array([[ELEVATOR_END_HEIGHT], [0.0]])
    )

    # Arm start and end conditions
    problem.subject_to(arm[:, :1] == np.array([[ARM_START_ANGLE], [0.0]]))
    problem.subject_to(arm[:, N : N + 1] == np.array([[ARM_END_ANGLE], [0.0]]))

    # Elevator velocity limits
    problem.subject_to(-ELEVATOR_MAX_VELOCITY <= elevator[1:2, :])
    problem.subject_to(elevator[1:2, :] <= ELEVATOR_MAX_VELOCITY)

    # Elevator acceleration limits
    problem.subject_to(-ELEVATOR_MAX_ACCELERATION <= elevator_accel)
    problem.subject_to(elevator_accel <= ELEVATOR_MAX_ACCELERATION)

    # Arm velocity limits
    problem.subject_to(-ARM_MAX_VELOCITY <= arm[1:2, :])
    problem.subject_to(arm[1:2, :] <= ARM_MAX_VELOCITY)

    # Arm acceleration limits
    problem.subject_to(-ARM_MAX_ACCELERATION <= arm_accel)
    problem.subject_to(arm_accel <= ARM_MAX_ACCELERATION)

    # Height limit
    if 0:
        heights = elevator[:1, :] + ARM_LENGTH * arm[:1, :].cwise_map(autodiff.sin)
        problem.subject_to(heights <= END_EFFECTOR_MAX_HEIGHT)

    # Cost function
    problem.minimize(
        sum(
            (ELEVATOR_END_HEIGHT - elevator[0, k]) ** 2
            + (ARM_END_ANGLE - arm[0, k]) ** 2
            for k in range(N + 1)
        )
    )

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS



import math

import numpy as np
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import VariableMatrix

# x = [x, y, heading, left velocity, right velocity]ᵀ
# u = [left voltage, right voltage]ᵀ
trackwidth = 0.699  # m
Kv_linear = 3.02  # V/(m/s)
Ka_linear = 0.642  # V/(m/s²)
Kv_angular = 1.382  # V/(m/s)
Ka_angular = 0.08495  # V/(m/s²)

A1 = -(Kv_linear / Ka_linear + Kv_angular / Ka_angular) / 2.0
A2 = -(Kv_linear / Ka_linear - Kv_angular / Ka_angular) / 2.0
B1 = 0.5 / Ka_linear + 0.5 / Ka_angular
B2 = 0.5 / Ka_linear - 0.5 / Ka_angular
A = np.array([[A1, A2], [A2, A1]])
B = np.array([[B1, B2], [B2, B1]])


def differential_drive_dynamics_double(x, u):
    xdot = np.empty((5, 1))

    v = (x[3, 0] + x[4, 0]) / 2.0
    xdot[0, :] = v * math.cos(x[2, 0])
    xdot[1, :] = v * math.sin(x[2, 0])
    xdot[2, :] = (x[4, 0] - x[3, 0]) / trackwidth
    xdot[3:5, :] = A @ x[3:5, :] + B @ u

    return xdot


def differential_drive_dynamics_variable(x, u):
    xdot = VariableMatrix(5, 1)

    v = (x[3, 0] + x[4, 0]) / 2.0
    xdot[0, :] = v * autodiff.cos(x[2, 0])
    xdot[1, :] = v * autodiff.sin(x[2, 0])
    xdot[2, :] = (x[4, 0] - x[3, 0]) / trackwidth
    xdot[3:5, :] = A @ x[3:5, :] + B @ u

    return xdot



import math

import numpy as np
import pytest
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import Gradient, Variable, VariableMatrix


def test_trivial_case():
    a = Variable()
    a.set_value(10)
    b = Variable()
    b.set_value(20)
    c = a

    assert Gradient(a, a).value()[0, 0] == 1.0
    assert Gradient(a, b).value()[0, 0] == 0.0
    assert Gradient(c, a).value()[0, 0] == 1.0
    assert Gradient(c, b).value()[0, 0] == 0.0


def test_unary_plus():
    a = Variable()
    a.set_value(10)
    c = +a

    assert c.value() == a.value()
    assert Gradient(c, a).value()[0, 0] == 1.0


def test_unary_minus():
    a = Variable()
    a.set_value(10)
    c = -a

    assert c.value() == -a.value()
    assert Gradient(c, a).value()[0, 0] == -1.0


def test_identical_variables():
    a = Variable()
    a.set_value(10)
    x = a
    c = a * a + x

    assert c.value() == a.value() * a.value() + x.value()
    assert Gradient(c, a).value()[0, 0] == 2 * a.value() + Gradient(x, a).value()[0, 0]
    assert (
        Gradient(c, x).value()[0, 0] == 2 * a.value() * Gradient(a, x).value()[0, 0] + 1
    )


def test_elementary():
    a = Variable()
    a.set_value(1.0)
    b = Variable()
    b.set_value(2.0)

    c = -2 * a
    assert Gradient(c, a).value()[0, 0] == -2.0

    c = a / 3.0
    assert Gradient(c, a).value()[0, 0] == 1.0 / 3.0

    a.set_value(100.0)
    b.set_value(200.0)

    c = a + b
    assert Gradient(c, a).value()[0, 0] == 1.0
    assert Gradient(c, b).value()[0, 0] == 1.0

    c = a - b
    assert Gradient(c, a).value()[0, 0] == 1.0
    assert Gradient(c, b).value()[0, 0] == -1.0

    c = -a + b
    assert Gradient(c, a).value()[0, 0] == -1.0
    assert Gradient(c, b).value()[0, 0] == 1.0

    c = a + 1
    assert Gradient(c, a).value()[0, 0] == 1.0


def test_comparison():
    x = Variable()
    x.set_value(10.0)
    a = Variable()
    a.set_value(10.0)
    b = Variable()
    b.set_value(200.0)

    assert a.value() == a.value()
    assert a.value() == x.value()
    assert a.value() == 10.0
    assert 10.0 == a.value()

    assert a.value() != b.value()
    assert a.value() != 20.0
    assert 20.0 != a.value()

    assert a.value() < b.value()
    assert a.value() < 20.0

    assert b.value() > a.value()
    assert 20.0 > a.value()

    assert a.value() <= a.value()
    assert a.value() <= x.value()
    assert a.value() <= b.value()
    assert a.value() <= 10.0
    assert a.value() <= 20.0

    assert a.value() >= a.value()
    assert x.value() >= a.value()
    assert b.value() >= a.value()
    assert 10.0 >= a.value()
    assert 20.0 >= a.value()

    # Comparison between variables and expressions
    assert a.value() == a.value() / a.value() * a.value()
    assert a.value() / a.value() * a.value() == a.value()

    assert a.value() != (a - a).value()
    assert (a - a).value() != a.value()

    assert (a - a).value() < a.value()
    assert a.value() < (a + a).value()

    assert (a + a).value() > a.value()
    assert a.value() > (a - a).value()

    assert a.value() <= (a - a + a).value()
    assert (a - a + a).value() <= a.value()

    assert a.value() <= (a + a).value()
    assert (a - a).value() <= a.value()

    assert a.value() >= (a - a + a).value()
    assert (a - a + a).value() >= a.value()

    assert (a + a).value() >= a.value()
    assert a.value() >= (a - a).value()


def test_trigonometry():
    x = Variable()
    x.set_value(0.5)

    # sin(x)
    assert autodiff.sin(x).value() == math.sin(x.value())

    g = Gradient(autodiff.sin(x), x)
    assert g.get().value()[0, 0] == math.cos(x.value())
    assert g.value()[0, 0] == math.cos(x.value())

    # cos(x)
    assert autodiff.cos(x).value() == math.cos(x.value())

    g = Gradient(autodiff.cos(x), x)
    assert g.get().value()[0, 0] == -math.sin(x.value())
    assert g.value()[0, 0] == -math.sin(x.value())

    # tan(x)
    assert autodiff.tan(x).value() == math.tan(x.value())

    g = Gradient(autodiff.tan(x), x)
    assert g.get().value()[0, 0] == 1.0 / (math.cos(x.value()) * math.cos(x.value()))
    assert g.value()[0, 0] == 1.0 / (math.cos(x.value()) * math.cos(x.value()))

    # asin(x)
    assert autodiff.asin(x).value() == math.asin(x.value())

    g = Gradient(autodiff.asin(x), x)
    assert g.get().value()[0, 0] == 1.0 / math.sqrt(1 - x.value() * x.value())
    assert g.value()[0, 0] == 1.0 / math.sqrt(1 - x.value() * x.value())

    # acos(x)
    assert autodiff.acos(x).value() == math.acos(x.value())

    g = Gradient(autodiff.acos(x), x)
    assert g.get().value()[0, 0] == -1.0 / math.sqrt(1 - x.value() * x.value())
    assert g.value()[0, 0] == -1.0 / math.sqrt(1 - x.value() * x.value())

    # atan(x)
    assert autodiff.atan(x).value() == math.atan(x.value())

    g = Gradient(autodiff.atan(x), x)
    assert g.get().value()[0, 0] == 1.0 / (1 + x.value() * x.value())
    assert g.value()[0, 0] == 1.0 / (1 + x.value() * x.value())


def test_hyperbolic():
    x = Variable()
    x.set_value(1.0)

    # sinh(x)
    assert autodiff.sinh(x).value() == math.sinh(x.value())

    g = Gradient(autodiff.sinh(x), x)
    assert g.get().value()[0, 0] == math.cosh(x.value())
    assert g.value()[0, 0] == math.cosh(x.value())

    # cosh(x)
    assert autodiff.cosh(x).value() == math.cosh(x.value())

    g = Gradient(autodiff.cosh(x), x)
    assert g.get().value()[0, 0] == math.sinh(x.value())
    assert g.value()[0, 0] == math.sinh(x.value())

    # tanh(x)
    assert autodiff.tanh(x).value() == math.tanh(x.value())

    g = Gradient(autodiff.tanh(x), x)
    assert g.get().value()[0, 0] == 1.0 / (math.cosh(x.value()) * math.cosh(x.value()))
    assert g.value()[0, 0] == 1.0 / (math.cosh(x.value()) * math.cosh(x.value()))


def test_exponential():
    x = Variable()
    x.set_value(1.0)

    # log(x)
    assert autodiff.log(x).value() == math.log(x.value())

    g = Gradient(autodiff.log(x), x)
    assert g.get().value()[0, 0] == 1.0 / x.value()
    assert g.value()[0, 0] == 1.0 / x.value()

    # log10(x)
    assert autodiff.log10(x).value() == math.log10(x.value())

    g = Gradient(autodiff.log10(x), x)
    assert g.get().value()[0, 0] == 1.0 / (math.log(10.0) * x.value())
    assert g.value()[0, 0] == 1.0 / (math.log(10.0) * x.value())

    # exp(x)
    assert autodiff.exp(x).value() == math.exp(x.value())

    g = Gradient(autodiff.exp(x), x)
    assert g.get().value()[0, 0] == math.exp(x.value())
    assert g.value()[0, 0] == math.exp(x.value())


def test_power():
    x = Variable()
    x.set_value(1.0)
    a = Variable()
    a.set_value(2.0)
    y = 2 * a

    # sqrt(x)
    assert autodiff.sqrt(x).value() == math.sqrt(x.value())

    g = Gradient(autodiff.sqrt(x), x)
    assert autodiff.sqrt(x).value() == math.sqrt(x.value())
    assert g.get().value()[0, 0] == 0.5 / math.sqrt(x.value())
    assert g.value()[0, 0] == 0.5 / math.sqrt(x.value())

    # sqrt(a)
    assert autodiff.sqrt(a).value() == math.sqrt(a.value())

    g = Gradient(autodiff.sqrt(a), a)
    assert autodiff.sqrt(a).value() == math.sqrt(a.value())
    assert g.get().value()[0, 0] == 0.5 / math.sqrt(a.value())
    assert g.value()[0, 0] == 0.5 / math.sqrt(a.value())

    # cbrt(x)
    assert autodiff.cbrt(x).value() == math.cbrt(x.value())

    g = Gradient(autodiff.cbrt(x), x)
    assert autodiff.cbrt(x).value() == math.cbrt(x.value())
    assert g.get().value()[0, 0] == 1.0 / (
        3.0 * math.cbrt(x.value()) * math.cbrt(x.value())
    )
    assert g.value()[0, 0] == 1.0 / (3.0 * math.cbrt(x.value()) * math.cbrt(x.value()))

    # cbrt(a)
    assert autodiff.cbrt(a).value() == math.cbrt(a.value())

    g = Gradient(autodiff.cbrt(a), a)
    assert autodiff.cbrt(a).value() == math.cbrt(a.value())
    assert g.get().value()[0, 0] == 1.0 / (
        3.0 * math.cbrt(a.value()) * math.cbrt(a.value())
    )
    assert g.value()[0, 0] == 1.0 / (3.0 * math.cbrt(a.value()) * math.cbrt(a.value()))

    # x²
    assert autodiff.pow(x, 2.0).value() == math.pow(x.value(), 2.0)

    g = Gradient(autodiff.pow(x, 2.0), x)
    assert g.get().value()[0, 0] == 2.0 * x.value()
    assert g.value()[0, 0] == 2.0 * x.value()

    # 2ˣ
    assert autodiff.pow(2.0, x).value() == math.pow(2.0, x.value())

    g = Gradient(autodiff.pow(2.0, x), x)
    assert g.get().value()[0, 0] == math.log(2.0) * math.pow(2.0, x.value())
    assert g.value()[0, 0] == math.log(2.0) * math.pow(2.0, x.value())

    # xˣ
    assert autodiff.pow(x, x).value() == math.pow(x.value(), x.value())

    g = Gradient(autodiff.pow(x, x), x)
    assert g.get().value()[0, 0] == (math.log(x.value()) + 1) * math.pow(
        x.value(), x.value()
    )
    assert g.value()[0, 0] == (math.log(x.value()) + 1) * math.pow(x.value(), x.value())

    # y(a)
    assert y.value() == 2 * a.value()

    g = Gradient(y, a)
    assert g.get().value()[0, 0] == 2.0
    assert g.value()[0, 0] == 2.0

    # xʸ(x)
    assert autodiff.pow(x, y).value() == math.pow(x.value(), y.value())

    g = Gradient(autodiff.pow(x, y), x)
    assert g.get().value()[0, 0] == y.value() / x.value() * math.pow(
        x.value(), y.value()
    )
    assert g.value()[0, 0] == y.value() / x.value() * math.pow(x.value(), y.value())

    # xʸ(a)
    assert autodiff.pow(x, y).value() == math.pow(x.value(), y.value())

    g = Gradient(autodiff.pow(x, y), a)
    assert g.get().value()[0, 0] == math.pow(x.value(), y.value()) * (
        y.value() / x.value() * Gradient(x, a).value()[0, 0]
        + math.log(x.value()) * Gradient(y, a).value()[0, 0]
    )
    assert g.value()[0, 0] == math.pow(x.value(), y.value()) * (
        y.value() / x.value() * Gradient(x, a).value()[0, 0]
        + math.log(x.value()) * Gradient(y, a).value()[0, 0]
    )

    # xʸ(y)
    assert autodiff.pow(x, y).value() == math.pow(x.value(), y.value())

    g = Gradient(autodiff.pow(x, y), y)
    assert g.get().value()[0, 0] == math.log(x.value()) * math.pow(x.value(), y.value())
    assert g.value()[0, 0] == math.log(x.value()) * math.pow(x.value(), y.value())


def test_abs():
    x = Variable()
    g = Gradient(autodiff.abs(x), x)

    x.set_value(1.0)
    assert autodiff.abs(x).value() == abs(x.value())
    assert g.get().value()[0, 0] == 1.0
    assert g.value()[0, 0] == 1.0

    x.set_value(-1.0)
    assert autodiff.abs(x).value() == abs(x.value())
    assert g.get().value()[0, 0] == -1.0
    assert g.value()[0, 0] == -1.0

    x.set_value(0.0)
    assert autodiff.abs(x).value() == abs(x.value())
    assert g.get().value()[0, 0] == 0.0
    assert g.value()[0, 0] == 0.0


def test_atan2():
    x = Variable()
    y = Variable()

    # Testing atan2 function on (double, var)
    x.set_value(1.0)
    y.set_value(0.9)
    assert autodiff.atan2(2.0, x).value() == math.atan2(2.0, x.value())

    g = Gradient(autodiff.atan2(2.0, x), x)
    assert g.get().value()[0, 0] == pytest.approx(
        -2.0 / (2 * 2 + x.value() * x.value()), abs=1e-15
    )
    assert g.value()[0, 0] == pytest.approx(
        -2.0 / (2 * 2 + x.value() * x.value()), abs=1e-15
    )

    # Testing atan2 function on (var, double)
    x.set_value(1.0)
    y.set_value(0.9)
    assert autodiff.atan2(x, 2.0).value() == math.atan2(x.value(), 2.0)

    g = Gradient(autodiff.atan2(x, 2.0), x)
    assert g.get().value()[0, 0] == pytest.approx(
        (2.0 / (2 * 2 + x * x)).value(), abs=1e-15
    )
    assert g.value()[0, 0] == pytest.approx(
        2.0 / (2 * 2 + x.value() * x.value()), abs=1e-15
    )

    # Testing atan2 function on (var, var)
    x.set_value(1.1)
    y.set_value(0.9)
    assert autodiff.atan2(y, x).value() == math.atan2(y.value(), x.value())

    g = Gradient(autodiff.atan2(y, x), y)
    assert g.get().value()[0, 0] == pytest.approx(
        (x / (x * x + y * y)).value(), abs=1e-15
    )
    assert g.value()[0, 0] == pytest.approx(
        x.value() / (x.value() * x.value() + y.value() * y.value()), abs=1e-15
    )

    g = Gradient(autodiff.atan2(y, x), x)
    assert g.get().value()[0, 0] == pytest.approx(
        -y.value() / (x.value() * x.value() + y.value() * y.value()), abs=1e-15
    )
    assert g.value()[0, 0] == pytest.approx(
        -y.value() / (x.value() * x.value() + y.value() * y.value()), abs=1e-15
    )

    # Testing atan2 function on (expr, expr)
    assert 3 * autodiff.atan2(autodiff.sin(y), 2 * x + 1).value() == 3 * math.atan2(
        math.sin(y.value()), 2 * x.value() + 1
    )

    g = Gradient(3 * autodiff.atan2(autodiff.sin(y), 2 * x + 1), y)
    assert g.get().value()[0, 0] == pytest.approx(
        3
        * (2 * x.value() + 1)
        * math.cos(y.value())
        / (
            (2 * x.value() + 1) * (2 * x.value() + 1)
            + math.sin(y.value()) * math.sin(y.value())
        ),
        abs=1e-15,
    )
    assert g.value()[0, 0] == pytest.approx(
        3
        * (2 * x.value() + 1)
        * math.cos(y.value())
        / (
            (2 * x.value() + 1) * (2 * x.value() + 1)
            + math.sin(y.value()) * math.sin(y.value())
        ),
        abs=1e-15,
    )

    g = Gradient(3 * autodiff.atan2(autodiff.sin(y), 2 * x + 1), x)
    assert g.get().value()[0, 0] == pytest.approx(
        3
        * -2
        * math.sin(y.value())
        / (
            (2 * x.value() + 1) * (2 * x.value() + 1)
            + math.sin(y.value()) * math.sin(y.value())
        ),
        abs=1e-15,
    )
    assert g.value()[0, 0] == pytest.approx(
        3
        * -2
        * math.sin(y.value())
        / (
            (2 * x.value() + 1) * (2 * x.value() + 1)
            + math.sin(y.value()) * math.sin(y.value())
        ),
        abs=1e-15,
    )


def test_hypot():
    x = Variable()
    y = Variable()

    # Testing hypot function on (var, double)
    x.set_value(1.8)
    y.set_value(1.5)
    assert autodiff.hypot(x, 2.0).value() == math.hypot(x.value(), 2.0)

    g = Gradient(autodiff.hypot(x, 2.0), x)
    assert g.get().value()[0, 0] == x.value() / math.hypot(x.value(), 2.0)
    assert g.value()[0, 0] == x.value() / math.hypot(x.value(), 2.0)

    # Testing hypot function on (double, var)
    assert autodiff.hypot(2.0, y).value() == math.hypot(2.0, y.value())

    g = Gradient(autodiff.hypot(2.0, y), y)
    assert g.get().value()[0, 0] == y.value() / math.hypot(2.0, y.value())
    assert g.value()[0, 0] == y.value() / math.hypot(2.0, y.value())

    # Testing hypot function on (var, var)
    x.set_value(1.3)
    y.set_value(2.3)
    assert autodiff.hypot(x, y).value() == math.hypot(x.value(), y.value())

    g = Gradient(autodiff.hypot(x, y), x)
    assert g.get().value()[0, 0] == x.value() / math.hypot(x.value(), y.value())
    assert g.value()[0, 0] == x.value() / math.hypot(x.value(), y.value())

    g = Gradient(autodiff.hypot(x, y), y)
    assert g.get().value()[0, 0] == y.value() / math.hypot(x.value(), y.value())
    assert g.value()[0, 0] == y.value() / math.hypot(x.value(), y.value())

    # Testing hypot function on (expr, expr)
    x.set_value(1.3)
    y.set_value(2.3)
    assert autodiff.hypot(2.0 * x, 3.0 * y).value() == math.hypot(
        2.0 * x.value(), 3.0 * y.value()
    )

    g = Gradient(autodiff.hypot(2.0 * x, 3.0 * y), x)
    assert g.get().value()[0, 0] == 4.0 * x.value() / math.hypot(
        2.0 * x.value(), 3.0 * y.value()
    )
    assert g.value()[0, 0] == 4.0 * x.value() / math.hypot(
        2.0 * x.value(), 3.0 * y.value()
    )

    g = Gradient(autodiff.hypot(2.0 * x, 3.0 * y), y)
    assert g.get().value()[0, 0] == 9.0 * y.value() / math.hypot(
        2.0 * x.value(), 3.0 * y.value()
    )
    assert g.value()[0, 0] == 9.0 * y.value() / math.hypot(
        2.0 * x.value(), 3.0 * y.value()
    )

    # Testing hypot function on (var, var, var)
    z = Variable()
    x.set_value(1.3)
    y.set_value(2.3)
    z.set_value(3.3)
    assert autodiff.hypot(x, y, z).value() == math.hypot(
        x.value(), y.value(), z.value()
    )

    g = Gradient(autodiff.hypot(x, y, z), x)
    assert g.get().value()[0, 0] == x.value() / math.hypot(
        x.value(), y.value(), z.value()
    )
    assert g.value()[0, 0] == x.value() / math.hypot(x.value(), y.value(), z.value())

    g = Gradient(autodiff.hypot(x, y, z), y)
    assert g.get().value()[0, 0] == y.value() / math.hypot(
        x.value(), y.value(), z.value()
    )
    assert g.value()[0, 0] == y.value() / math.hypot(x.value(), y.value(), z.value())

    g = Gradient(autodiff.hypot(x, y, z), z)
    assert g.get().value()[0, 0] == z.value() / math.hypot(
        x.value(), y.value(), z.value()
    )
    assert g.value()[0, 0] == z.value() / math.hypot(x.value(), y.value(), z.value())


def test_miscellaneous():
    x = Variable()

    # dx/dx
    x.set_value(3.0)
    assert autodiff.abs(x).value() == abs(x.value())

    g = Gradient(x, x)
    assert g.get().value()[0, 0] == 1.0
    assert g.value()[0, 0] == 1.0

    # erf(x)
    x.set_value(0.5)
    assert autodiff.erf(x).value() == math.erf(x.value())

    g = Gradient(autodiff.erf(x), x)
    assert g.get().value()[0, 0] == 2.0 / math.sqrt(math.pi) * math.exp(
        -x.value() * x.value()
    )
    assert g.value()[0, 0] == 2.0 / math.sqrt(math.pi) * math.exp(
        -x.value() * x.value()
    )


def test_reuse():
    a = Variable()
    a.set_value(10)

    b = Variable()
    b.set_value(20)

    x = a * b

    g = Gradient(x, a)

    assert g.get().value()[0] == 20.0
    assert g.value()[0] == 20.0

    b.set_value(10)
    assert g.get().value()[0] == 10.0
    assert g.value()[0] == 10.0


def test_sign():
    def sign(x):
        if x < 0.0:
            return -1.0
        elif x == 0.0:
            return 0.0
        else:
            return 1.0

    x = Variable()

    # sign(1.0)
    x.set_value(1.0)
    assert sign(x.value()) == autodiff.sign(x).value()

    g = Gradient(autodiff.sign(x), x)
    assert g.get().value()[0, 0] == 0.0
    assert g.value()[0, 0] == 0.0

    # sign(-1.0)
    x.set_value(-1.0)
    assert sign(x.value()) == autodiff.sign(x).value()

    g = Gradient(autodiff.sign(x), x)
    assert g.get().value()[0, 0] == 0.0
    assert g.value()[0, 0] == 0.0

    # sign(0.0)
    x.set_value(0.0)
    assert sign(x.value()) == autodiff.sign(x).value()

    g = Gradient(autodiff.sign(x), x)
    assert g.get().value()[0, 0] == 0.0
    assert g.value()[0, 0] == 0.0


def test_non_scalar():
    x = VariableMatrix(3)
    x[0].set_value(1)
    x[1].set_value(2)
    x[2].set_value(3)

    # y = [x₁ + 3x₂ − 5x₃]
    #
    # dy/dx = [1  3  −5]
    y = x[0] + 3 * x[1] - 5 * x[2]
    g = Gradient(y, x)

    expected_g = np.array([[1.0], [3.0], [-5.0]])

    g_get_value = g.get().value()
    assert g_get_value.shape == (3, 1)
    assert (g_get_value == expected_g).all()

    g_value = g.get().value()
    assert g_value.shape == (3, 1)
    assert (g_value == expected_g).all()



import math

import numpy as np
import pytest
from cart_pole_util import (
    cart_pole_dynamics_double,
    cart_pole_dynamics_variable,
)
from rk4 import rk4
from sleipnir.autodiff import ExpressionType, VariableMatrix
from sleipnir.optimization import (
    OCP,
    DynamicsType,
    ExitStatus,
    TimestepMethod,
    TranscriptionMethod,
)


def lerp(a, b, t):
    return a + t * (b - a)


def test_cart_pole_ocp():
    TOTAL_TIME = 5  # s
    dt = 0.05  # s
    N = int(TOTAL_TIME / dt)

    u_max = 20.0  # N
    d_max = 2.0  # m

    x_initial = np.zeros((4, 1))
    x_final = np.array([[1.0], [math.pi], [0.0], [0.0]])

    problem = OCP(
        4,
        1,
        dt,
        N,
        cart_pole_dynamics_variable,
        DynamicsType.EXPLICIT_ODE,
        TimestepMethod.VARIABLE_SINGLE,
        TranscriptionMethod.DIRECT_COLLOCATION,
    )

    # x = [q, q̇]ᵀ = [x, θ, ẋ, θ̇]ᵀ
    X = problem.X()

    # Initial guess
    for k in range(N + 1):
        X[0, k].set_value(lerp(x_initial[0, 0], x_final[0, 0], k / N))
        X[1, k].set_value(lerp(x_initial[1, 0], x_final[1, 0], k / N))

    # Initial conditions
    problem.constrain_initial_state(x_initial)

    # Final conditions
    problem.constrain_final_state(x_final)

    # Cart position constraints
    def each(x: VariableMatrix, u: VariableMatrix):
        problem.subject_to(x[0] >= 0.0)
        problem.subject_to(x[0] <= d_max)

    problem.for_each_step(each)

    # Input constraints
    problem.set_lower_input_bound(-u_max)
    problem.set_upper_input_bound(u_max)

    # u = f_x
    U = problem.U()

    # Minimize sum squared inputs
    problem.minimize(sum(U[:, k : k + 1].T @ U[:, k : k + 1] for k in range(N)))

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONLINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    # Verify initial state
    assert X.value(0, 0) == pytest.approx(x_initial[0, 0], abs=1e-8)
    assert X.value(1, 0) == pytest.approx(x_initial[1, 0], abs=1e-8)
    assert X.value(2, 0) == pytest.approx(x_initial[2, 0], abs=1e-8)
    assert X.value(3, 0) == pytest.approx(x_initial[3, 0], abs=1e-8)

    # FIXME: Replay diverges
    return

    # Verify solution
    x = np.zeros((4, 1))
    u = np.zeros((1, 1))
    for k in range(N):
        # Cart position constraints
        assert X[0, k] >= 0.0
        assert X[0, k] <= d_max

        # Input constraints
        assert U[0, k] >= -u_max
        assert U[0, k] <= u_max

        # Verify state
        assert X.value(0, k) == pytest.approx(x[0, 0], abs=1e-2)
        assert X.value(1, k) == pytest.approx(x[1, 0], abs=1e-2)
        assert X.value(2, k) == pytest.approx(x[2, 0], abs=1e-2)
        assert X.value(3, k) == pytest.approx(x[3, 0], abs=1e-2)

        # Project state forward
        x = rk4(cart_pole_dynamics_double, x, u, dt)

    # Verify final state
    assert X.value(0, N) == pytest.approx(x_final[0, 0], abs=1e-8)
    assert X.value(1, N) == pytest.approx(x_final[1, 0], abs=1e-8)
    assert X.value(2, N) == pytest.approx(x_final[2, 0], abs=1e-8)
    assert X.value(3, N) == pytest.approx(x_final[3, 0], abs=1e-8)

    # Log states for offline viewing
    with open("Cart-pole states.csv", "w") as f:
        f.write(
            "Time (s),Cart position (m),Pole angle (rad),Cart velocity (m/s),Pole angular velocity (rad/s)\n"
        )

        for k in range(N + 1):
            f.write(
                f"{k * dt},{X.value(0, k)},{X.value(1, k)},{X.value(2, k)},{X.value(3, k)}\n"
            )

    # Log inputs for offline viewing
    with open("Cart-pole inputs.csv", "w") as f:
        f.write("Time (s),Cart force (N)\n")

        for k in range(N + 1):
            if k < N:
                f.write(f"{k * dt},{U.value(0, k)}\n")
            else:
                f.write(f"{k * dt},0.0\n")



"""These tests ensure coverage of the off-nominal exit statuses"""

from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_callback_requested_stop():
    problem = Problem()

    x = problem.decision_variable()
    problem.minimize(x * x)

    problem.add_callback(lambda info: False)
    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    problem.add_callback(lambda info: True)
    assert problem.solve(diagnostics=True) == ExitStatus.CALLBACK_REQUESTED_STOP

    problem.clear_callbacks()
    problem.add_callback(lambda info: False)
    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS


def test_too_few_dofs():
    problem = Problem()

    x = problem.decision_variable()
    y = problem.decision_variable()
    z = problem.decision_variable()

    problem.subject_to(x == 1)
    problem.subject_to(x == 2)
    problem.subject_to(y == 1)
    problem.subject_to(z == 1)

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.TOO_FEW_DOFS


def test_locally_infeasible_equality_constraints():
    problem = Problem()

    x = problem.decision_variable()
    y = problem.decision_variable()
    z = problem.decision_variable()

    problem.subject_to(x == y + 1)
    problem.subject_to(y == z + 1)
    problem.subject_to(z == x + 1)

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.LOCALLY_INFEASIBLE


def test_locally_infeasible_inequality_constraints():
    problem = Problem()

    x = problem.decision_variable()
    y = problem.decision_variable()
    z = problem.decision_variable()

    problem.subject_to(x >= y + 1)
    problem.subject_to(y >= z + 1)
    problem.subject_to(z >= x + 1)

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.LOCALLY_INFEASIBLE


def test_diverging_iterates():
    problem = Problem()

    x = problem.decision_variable()
    problem.minimize(x)

    assert problem.cost_function_type() == ExpressionType.LINEAR
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.DIVERGING_ITERATES


def test_max_iterations_exceeded():
    problem = Problem()

    x = problem.decision_variable()
    problem.minimize(x * x)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert (
        problem.solve(max_iterations=0, diagnostics=True)
        == ExitStatus.MAX_ITERATIONS_EXCEEDED
    )


def test_timeout():
    problem = Problem()

    x = problem.decision_variable()
    problem.minimize(x * x)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(timeout=0.0, diagnostics=True) == ExitStatus.TIMEOUT



import numpy as np
from sleipnir.optimization import Problem


def test_scalar_init_assign():
    problem = Problem()

    # Scalar zero init
    x = problem.decision_variable()
    assert x.value() == 0.0

    # Scalar assignment
    x.set_value(1.0)
    assert x.value() == 1.0
    x.set_value(2.0)
    assert x.value() == 2.0


def test_vector_init_assign():
    problem = Problem()

    # Vector zero init
    y = problem.decision_variable(2)
    assert y.value(0) == 0.0
    assert y.value(1) == 0.0

    # Vector assignment
    y[0].set_value(1.0)
    y[1].set_value(2.0)
    assert y.value(0) == 1.0
    assert y.value(1) == 2.0
    y[0].set_value(3.0)
    y[1].set_value(4.0)
    assert y.value(0) == 3.0
    assert y.value(1) == 4.0


def test_matrix_init_assign():
    problem = Problem()

    # Matrix zero init
    z = problem.decision_variable(3, 2)
    assert z.value(0, 0) == 0.0
    assert z.value(0, 1) == 0.0
    assert z.value(1, 0) == 0.0
    assert z.value(1, 1) == 0.0
    assert z.value(2, 0) == 0.0
    assert z.value(2, 1) == 0.0

    # Matrix assignment; element comparison
    z.set_value(np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]]))
    assert z.value(0, 0) == 1.0
    assert z.value(0, 1) == 2.0
    assert z.value(1, 0) == 3.0
    assert z.value(1, 1) == 4.0
    assert z.value(2, 0) == 5.0
    assert z.value(2, 1) == 6.0

    # Matrix assignment; matrix comparison
    z.set_value(np.array([[7.0, 8.0], [9.0, 10.0], [11.0, 12.0]]))
    assert (z.value() == np.array([[7.0, 8.0], [9.0, 10.0], [11.0, 12.0]])).all()

    # Block assignment
    z[:2, :1].set_value(np.array([[1.0], [1.0]]))
    assert (z.value() == np.array([[1.0, 8.0], [1.0, 10.0], [11.0, 12.0]])).all()


def test_symmetric_matrix():
    problem = Problem()

    # Matrix zero init
    A = problem.symmetric_decision_variable(2)
    assert A.value(0, 0) == 0.0
    assert A.value(0, 1) == 0.0
    assert A.value(1, 0) == 0.0
    assert A.value(1, 1) == 0.0

    # Assign to lower triangle
    A[0, 0].set_value(1.0)
    A[1, 0].set_value(2.0)
    A[1, 1].set_value(3.0)

    # Confirm whole matrix changed
    assert A.value(0, 0) == 1.0
    assert A.value(0, 1) == 2.0
    assert A.value(1, 0) == 2.0
    assert A.value(1, 1) == 3.0



import numpy as np
import pytest
from differential_drive_util import (
    differential_drive_dynamics_double,
    differential_drive_dynamics_variable,
)
from rk4 import rk4
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def lerp(a, b, t):
    return a + t * (b - a)


def test_differential_drive_problem():
    TOTAL_TIME = 5.0  # s
    dt = 0.05  # s
    N = int(TOTAL_TIME / dt)

    u_max = 12.0  # V

    x_initial = np.zeros((5, 1))
    x_final = np.array([[1.0], [1.0], [0.0], [0.0], [0.0]])

    problem = Problem()

    # x = [x, y, heading, left velocity, right velocity]ᵀ
    X = problem.decision_variable(5, N + 1)

    # Initial guess
    for k in range(N + 1):
        X[0, k].set_value(lerp(x_initial[0, 0], x_final[0, 0], k / N))
        X[1, k].set_value(lerp(x_initial[1, 0], x_final[1, 0], k / N))

    # u = [left voltage, right voltage]ᵀ
    U = problem.decision_variable(2, N)

    # Initial conditions
    problem.subject_to(X[:, :1] == x_initial)

    # Final conditions
    problem.subject_to(X[:, N : N + 1] == x_final)

    # Input constraints
    problem.subject_to(U >= -u_max)
    problem.subject_to(U <= u_max)

    # Dynamics constraints - RK4 integration
    for k in range(N):
        problem.subject_to(
            X[:, k + 1 : k + 2]
            == rk4(
                differential_drive_dynamics_variable,
                X[:, k : k + 1],
                U[:, k : k + 1],
                dt,
            )
        )

    # Minimize sum squared states and inputs
    problem.minimize(
        sum(
            X[:, k : k + 1].T @ X[:, k : k + 1] + U[:, k : k + 1].T @ U[:, k : k + 1]
            for k in range(N)
        )
    )

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.NONLINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    # Verify initial state
    assert X.value(0, 0) == pytest.approx(x_initial[0, 0], abs=1e-8)
    assert X.value(1, 0) == pytest.approx(x_initial[1, 0], abs=1e-8)
    assert X.value(2, 0) == pytest.approx(x_initial[2, 0], abs=1e-8)
    assert X.value(3, 0) == pytest.approx(x_initial[3, 0], abs=1e-8)
    assert X.value(4, 0) == pytest.approx(x_initial[4, 0], abs=1e-8)

    # Verify solution
    x = np.zeros((5, 1))
    for k in range(N):
        u = U[:, k : k + 1].value()

        # Input constraints
        assert U[0, k].value() >= -u_max
        assert U[0, k].value() <= u_max
        assert U[1, k].value() >= -u_max
        assert U[1, k].value() <= u_max

        # Verify state
        assert X.value(0, k) == pytest.approx(x[0, 0], abs=1e-8)
        assert X.value(1, k) == pytest.approx(x[1, 0], abs=1e-8)
        assert X.value(2, k) == pytest.approx(x[2, 0], abs=1e-8)
        assert X.value(3, k) == pytest.approx(x[3, 0], abs=1e-8)
        assert X.value(4, k) == pytest.approx(x[4, 0], abs=1e-8)

        # Project state forward
        x = rk4(differential_drive_dynamics_double, x, u, dt)

    # Verify final state
    assert X.value(0, N) == pytest.approx(x_final[0, 0], abs=1e-8)
    assert X.value(1, N) == pytest.approx(x_final[1, 0], abs=1e-8)
    assert X.value(2, N) == pytest.approx(x_final[2, 0], abs=1e-8)
    assert X.value(3, N) == pytest.approx(x_final[3, 0], abs=1e-8)
    assert X.value(4, N) == pytest.approx(x_final[4, 0], abs=1e-8)

    # Log states for offline viewing
    with open("Differential drive states.csv", "w") as f:
        f.write(
            "Time (s),X position (m),Y position (m),Heading (rad),Left velocity (m/s),Right velocity (m/s)\n"
        )

        for k in range(N + 1):
            f.write(
                f"{k * dt},{X.value(0, k)},{X.value(1, k)},{X.value(2, k)},{X.value(3, k)},{X.value(4, k)}\n"
            )

    # Log inputs for offline viewing
    with open("Differential drive inputs.csv", "w") as f:
        f.write("Time (s),Left voltage (V),Right voltage (V)\n")

        for k in range(N + 1):
            if k < N:
                f.write(f"{k * dt},{U.value(0, k)},{U.value(1, k)}\n")
            else:
                f.write(f"{k * dt},0.0,0.0\n")



import pytest
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_maximize():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(1.0)

    y = problem.decision_variable()
    y.set_value(1.0)

    problem.maximize(50 * x + 40 * y)

    problem.subject_to(x + 1.5 * y <= 750)
    problem.subject_to(2 * x + 3 * y <= 1500)
    problem.subject_to(2 * x + y <= 1000)
    problem.subject_to(x >= 0)
    problem.subject_to(y >= 0)

    assert problem.cost_function_type() == ExpressionType.LINEAR
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(375.0, abs=1e-6)
    assert y.value() == pytest.approx(250.0, abs=1e-6)


def test_free_variable():
    problem = Problem()

    x = problem.decision_variable(2)
    x[0].set_value(1.0)
    x[1].set_value(2.0)

    problem.subject_to(x[0] == 0)

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x[0].value() == pytest.approx(0.0, abs=1e-6)
    assert x[1].value() == pytest.approx(2.0, abs=1e-6)



import pytest
import sleipnir.autodiff as autodiff
from sleipnir.optimization import ExitStatus, Problem, multistart


class DecisionVariables:
    def __init__(self, x, y):
        self.x = x
        self.y = y


def mishras_bird_function_solve(input: DecisionVariables):
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(input.x)
    y = problem.decision_variable()
    y.set_value(input.y)

    # https://en.wikipedia.org/wiki/Test_functions_for_optimization#Test_functions_for_constrained_optimization
    J = (
        autodiff.sin(y) * autodiff.exp((1 - autodiff.cos(x)) ** 2)
        + autodiff.cos(x) * autodiff.exp((1 - autodiff.sin(y)) ** 2)
        + (x - y) ** 2
    )
    problem.minimize(J)

    problem.subject_to((x + 5) ** 2 + (y + 5) ** 2 < 25)

    return problem.solve(), J.value(), DecisionVariables(x.value(), y.value())


def test_mishras_bird_function():
    status, cost, variables = multistart(
        mishras_bird_function_solve,
        [DecisionVariables(-3, -8), DecisionVariables(-3, -1.5)],
    )

    assert status == ExitStatus.SUCCESS

    assert variables.x == pytest.approx(-3.130246803458174, abs=1e-15)
    assert variables.y == pytest.approx(-1.5821421769364057, abs=1e-15)



def rk4(f, x, u, dt):
    """
    Performs 4th order Runge-Kutta integration of dx/dt = f(x, u) for dt.

    Parameter ``f``:
        The function to integrate. It must take two arguments x and u.

    Parameter ``x``:
        The initial value of x.

    Parameter ``u``:
        The value u held constant over the integration period.

    Parameter ``dt``:
        The time over which to integrate.
    """
    h = dt

    k1 = f(x, u)
    k2 = f(x + h * 0.5 * k1, u)
    k3 = f(x + h * 0.5 * k2, u)
    k4 = f(x + h * k3, u)

    return x + h / 6.0 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)



import math

import numpy as np
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import Variable, VariableMatrix

# https://underactuated.mit.edu/acrobot.html#cart_pole
#
# θ is CCW+ measured from negative y-axis.
#
# q = [x, θ]ᵀ
# q̇ = [ẋ, θ̇]ᵀ
# u = f_x
#
# M(q)q̈ + C(q, q̇)q̇ = τ_g(q) + Bu
# M(q)q̈ = τ_g(q) − C(q, q̇)q̇ + Bu
# q̈ = M⁻¹(q)(τ_g(q) − C(q, q̇)q̇ + Bu)
#
#        [ m_c + m_p  m_p l cosθ]
# M(q) = [m_p l cosθ    m_p l²  ]
#
#           [0  −m_p lθ̇ sinθ]
# C(q, q̇) = [0       0      ]
#
#          [     0      ]
# τ_g(q) = [-m_p gl sinθ]
#
#     [1]
# B = [0]

m_c = 5.0  # Cart mass (kg)
m_p = 0.5  # Pole mass (kg)
l = 0.5  # Pole length (m)
g = 9.806  # Acceleration due to gravity (m/s²)


def cart_pole_dynamics_double(x, u):
    q = x[:2, :]
    qdot = x[2:, :]
    theta = q[1, 0]
    thetadot = qdot[1, 0]

    #        [ m_c + m_p  m_p l cosθ]
    # M(q) = [m_p l cosθ    m_p l²  ]
    M = np.array(
        [
            [m_c + m_p, m_p * l * math.cos(theta)],
            [m_p * l * math.cos(theta), m_p * l**2],
        ]
    )

    #           [0  −m_p lθ̇ sinθ]
    # C(q, q̇) = [0       0      ]
    C = np.array([[0.0, -m_p * l * thetadot * math.sin(theta)], [0.0, 0.0]])

    #          [     0      ]
    # τ_g(q) = [-m_p gl sinθ]
    tau_g = np.array([[0.0], [-m_p * g * l * math.sin(theta)]])

    #     [1]
    # B = [0]
    B = np.array([[1], [0]])

    # q̈ = M⁻¹(q)(τ_g(q) − C(q, q̇)q̇ + Bu)
    qddot = np.empty((4, 1))
    qddot[:2, :] = qdot
    qddot[2:, :] = np.linalg.solve(M, tau_g - C @ qdot + B @ u)
    return qddot


def cart_pole_dynamics_variable(x, u):
    q = x[:2, :]
    qdot = x[2:, :]
    theta = q[1, 0]
    thetadot = qdot[1, 0]

    #        [ m_c + m_p  m_p l cosθ]
    # M(q) = [m_p l cosθ    m_p l²  ]
    M = VariableMatrix(
        [
            [Variable(m_c + m_p), m_p * l * autodiff.cos(theta)],
            [m_p * l * autodiff.cos(theta), Variable(m_p * l**2)],
        ]
    )

    #           [0  −m_p lθ̇ sinθ]
    # C(q, q̇) = [0       0      ]
    C = VariableMatrix(
        [
            [Variable(0.0), -m_p * l * thetadot * autodiff.sin(theta)],
            [Variable(0.0), Variable(0.0)],
        ]
    )

    #          [     0      ]
    # τ_g(q) = [-m_p gl sinθ]
    tau_g = VariableMatrix([[Variable(0.0)], [-m_p * g * l * autodiff.sin(theta)]])

    #     [1]
    # B = [0]
    B = np.array([[1], [0]])

    # q̈ = M⁻¹(q)(τ_g(q) − C(q, q̇)q̇ + Bu)
    qddot = VariableMatrix(4, 1)
    qddot[:2, :] = qdot
    qddot[2:, :] = autodiff.solve(M, tau_g - C @ qdot + B @ u)
    return qddot



import numpy as np
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_empty():
    problem = Problem()

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS


def test_no_cost_unconstrained_1():
    problem = Problem()

    X = problem.decision_variable(2, 3)

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    for row in range(X.rows()):
        for col in range(X.cols()):
            assert X.value(row, col) == 0.0


def test_no_cost_unconstrained_2():
    problem = Problem()

    X = problem.decision_variable(2, 3)
    X.set_value(np.ones((2, 3)))

    assert problem.cost_function_type() == ExpressionType.NONE
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    for row in range(X.rows()):
        for col in range(X.cols()):
            assert X.value(row, col) == 1.0



import numpy as np
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import VariableMatrix


def test_assignment_to_default():
    mat = VariableMatrix()

    assert mat.rows() == 0
    assert mat.cols() == 0
    assert mat.shape == (0, 0)

    mat = VariableMatrix(2, 2)

    assert mat.rows() == 2
    assert mat.cols() == 2
    assert mat.shape == (2, 2)
    assert mat[0, 0].value() == 0.0
    assert mat[0, 1].value() == 0.0
    assert mat[1, 0].value() == 0.0
    assert mat[1, 1].value() == 0.0

    mat[0, 0].set_value(1.0)
    mat[0, 1].set_value(2.0)
    mat[1, 0].set_value(3.0)
    mat[1, 1].set_value(4.0)

    assert mat[0, 0].value() == 1.0
    assert mat[0, 1].value() == 2.0
    assert mat[1, 0].value() == 3.0
    assert mat[1, 1].value() == 4.0


def test_assignment_aliasing():
    A = VariableMatrix([[1.0, 2.0], [3.0, 4.0]])
    B = VariableMatrix([[5.0, 6.0], [7.0, 8.0]])

    # A and B initially contain different values
    expected_A = np.array([[1.0, 2.0], [3.0, 4.0]])
    expected_B = np.array([[5.0, 6.0], [7.0, 8.0]])
    assert (A.value() == expected_A).all()
    assert (B.value() == expected_B).all()

    # Make A point to B's storage
    A = B
    assert (A.value() == expected_B).all()
    assert (B.value() == expected_B).all()

    # Changes to B should be reflected in A
    B[0, 0].set_value(2.0)
    expected_B[0, 0] = 2.0
    assert (A.value() == expected_B).all()
    assert (B.value() == expected_B).all()


def test_slicing():
    mat = VariableMatrix(
        [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    )
    assert mat.shape == (4, 4)

    # Single-arg index operator on full matrix
    for i in range(mat.shape[0] * mat.shape[1]):
        assert mat[i] == i + 1

    # Slice from start
    s = mat[1:, 2:]
    assert s.shape == (3, 2)
    # Single-arg index operator on forward slice
    assert s[0] == 7.0
    assert s[1] == 8.0
    assert s[2] == 11.0
    assert s[3] == 12.0
    assert s[4] == 15.0
    assert s[5] == 16.0
    # Double-arg index operator on forward slice
    assert s[0, 0] == 7.0
    assert s[0, 1] == 8.0
    assert s[1, 0] == 11.0
    assert s[1, 1] == 12.0
    assert s[2, 0] == 15.0
    assert s[2, 1] == 16.0

    # Slice from end
    s = mat[-1:, -2:]
    assert s.shape == (1, 2)
    # Single-arg index operator on reverse slice
    assert s[0] == 15.0
    assert s[1] == 16.0
    # Double-arg index operator on reverse slice
    assert s[0, 0] == 15.0
    assert s[0, 1] == 16.0

    # Slice from start with step of 2
    s = mat[:, ::2]
    assert s.shape == (4, 2)
    assert (
        s.value() == np.array([[1.0, 3.0], [5.0, 7.0], [9.0, 11.0], [13.0, 15.0]])
    ).all()

    # Slice from end with negative step for row and column
    s = mat[::-1, ::-2]
    assert s.shape == (4, 2)
    assert (
        s.value()
        == np.array(
            [
                [16.0, 14.0],
                [12.0, 10.0],
                [8.0, 6.0],
                [4.0, 2.0],
            ]
        )
    ).all()

    # Slice from start and column -1
    s = mat[1:, -1]
    assert s.shape == (3, 1)
    assert (s.value() == np.array([[8.0], [12.0], [16.0]])).all()

    # Slice from start and column -2
    s = mat[1:, -2]
    assert s.shape == (3, 1)
    assert (s.value() == np.array([[7.0], [11.0], [15.0]])).all()

    # Block assignment
    assert mat[::2, ::2].shape == (2, 2)
    mat[::2, ::2] = np.array([[17.0, 18.0], [19.0, 20.0]])
    assert (
        mat.value()
        == np.array(
            [
                [17.0, 2.0, 18.0, 4.0],
                [5.0, 6.0, 7.0, 8.0],
                [19.0, 10.0, 20.0, 12.0],
                [13.0, 14.0, 15.0, 16.0],
            ]
        )
    ).all()


def test_subslicing():
    # Block-of-block assignment (row skip forward)
    mat = VariableMatrix(5, 5)
    assert mat[::2, ::1][1:3, 1:4].shape == (2, 3)
    mat[::2, ::1][1:3, 1:4] = np.array([[1, 2, 3], [4, 5, 6]])

    assert (
        mat.value()
        == np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 1, 2, 3, 0],
                [0, 0, 0, 0, 0],
                [0, 4, 5, 6, 0],
            ]
        )
    ).all()

    # Block-of-block assignment (row skip backward)
    mat = VariableMatrix(5, 5)
    assert mat[::-2, ::-1][1:3, 1:4].shape == (2, 3)
    mat[::-2, ::-1][1:3, 1:4] = np.array([[1, 2, 3], [4, 5, 6]])

    assert (
        mat.value()
        == np.array(
            [
                [0, 6, 5, 4, 0],
                [0, 0, 0, 0, 0],
                [0, 3, 2, 1, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    ).all()

    # Block-of-block assignment (column skip forward)
    mat = VariableMatrix(5, 5)
    assert mat[::1, ::2][1:4, 1:3].shape == (3, 2)
    mat[::1, ::2][1:4, 1:3] = np.array([[1, 2], [3, 4], [5, 6]])

    assert (
        mat.value()
        == np.array(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 1, 0, 2],
                [0, 0, 3, 0, 4],
                [0, 0, 5, 0, 6],
                [0, 0, 0, 0, 0],
            ]
        )
    ).all()

    # Block-of-block assignment (column skip backward)
    mat = VariableMatrix(5, 5)
    assert mat[::-1, ::-2][1:4, 1:3].shape == (3, 2)
    mat[::-1, ::-2][1:4, 1:3] = np.array([[1, 2], [3, 4], [5, 6]])

    assert (
        mat.value()
        == np.array(
            [
                [0, 0, 0, 0, 0],
                [6, 0, 5, 0, 0],
                [4, 0, 3, 0, 0],
                [2, 0, 1, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    ).all()


def test_iterators():
    A = VariableMatrix([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
    sub_A = A[2:3, 1:3]

    assert sum(1 for _ in A) == 9
    assert sum(1 for _ in sub_A) == 2

    i = 1
    for elem in A:
        assert elem.value() == i
        i += 1

    i = 8
    for elem in sub_A:
        assert elem.value() == i
        i += 1


def test_value():
    A = VariableMatrix([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
    expected = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])

    # Full matrix
    assert (A.value() == expected).all()
    assert A.value(3) == 4.0
    assert A.T.value(3) == 2.0

    # Slice
    assert (A[1:3, 1:3].value() == expected[1:3, 1:3]).all()
    assert A[1:3, 1:3].value(2) == 8.0
    assert A[1:3, 1:3].T.value(2) == 6.0

    # Slice-of-slice
    assert (A[1:3, 1:3][:, 1:].value() == expected[1:3, 1:3][:, 1:]).all()
    assert A[1:3, 1:3][:, 1:].value(1) == 9.0
    assert A[1:3, 1:3].T[:, 1:].value(1) == 9.0


def test_pow():
    # VariableMatrix coefficient-wise power
    A = VariableMatrix([[-2.0, -3.0, -4.0], [-5.0, -6.0, -7.0]])
    assert A**2 == np.array([[4.0, 9.0, 16.0], [25.0, 36.0, 49.0]])
    assert A**3 == np.array([[-8.0, -27.0, -64.0], [-125.0, -216.0, -343.0]])

    # VariableBlock coefficient-wise power
    sub_A = A[:2, :2]
    assert sub_A**2 == np.array([[4.0, 9.0], [25.0, 36.0]])
    assert sub_A**3 == np.array([[-8.0, -27.0], [-125.0, -216.0]])


def test_cwise_map():
    # VariableMatrix cwise_map
    A = VariableMatrix([[-2.0, -3.0, -4.0], [-5.0, -6.0, -7.0]])

    result1 = A.cwise_map(autodiff.abs)
    expected1 = np.array([[2.0, 3.0, 4.0], [5.0, 6.0, 7.0]])

    # Don't modify original matrix
    assert (A.value() == -expected1).all()

    assert (result1.value() == expected1).all()

    # VariableBlock cwise_map
    sub_A = A[:2, :2]

    result2 = sub_A.cwise_map(autodiff.abs)
    expected2 = np.array([[2.0, 3.0], [5.0, 6.0]])

    # Don't modify original matrix
    assert (A.value() == -expected1).all()
    assert (sub_A.value() == -expected2).all()

    assert (result2.value() == expected2).all()


def test_zero_static_function():
    A = VariableMatrix.zero(2, 3)

    for row in range(A.rows()):
        for col in range(A.cols()):
            assert A[row, col].value() == 0.0


def test_ones_static_function():
    A = VariableMatrix.ones(2, 3)

    for row in range(A.rows()):
        for col in range(A.cols()):
            assert A[row, col].value() == 1.0


def test_cwise_reduce():
    A = VariableMatrix([[2.0, 3.0, 4.0], [5.0, 6.0, 7.0]])
    B = VariableMatrix([[8.0, 9.0, 10.0], [11.0, 12.0, 13.0]])
    result = autodiff.cwise_reduce(A, B, lambda a, b: a * b)

    assert (result.value() == np.array([[16.0, 27.0, 40.0], [55.0, 72.0, 91.0]])).all()


def test_block_free_function():
    A = VariableMatrix([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    B = VariableMatrix([[7.0], [8.0]])

    mat1 = autodiff.block([[A, B]])
    expected1 = np.array([[1.0, 2.0, 3.0, 7.0], [4.0, 5.0, 6.0, 8.0]])
    assert mat1.shape == (2, 4)
    assert (mat1.value() == expected1).all()

    C = VariableMatrix([[9.0, 10.0, 11.0, 12.0]])

    mat2 = autodiff.block([[A, B], [C]])
    expected2 = np.array(
        [[1.0, 2.0, 3.0, 7.0], [4.0, 5.0, 6.0, 8.0], [9.0, 10.0, 11.0, 12.0]]
    )
    assert mat2.shape == (3, 4)
    assert (mat2.value() == expected2).all()


def check_solve(A: VariableMatrix, B: VariableMatrix):
    print(f"Solve {A.shape[0]}x{A.shape[1]}")

    X = autodiff.solve(A, B)

    assert X.shape == (A.shape[1], B.shape[1])
    assert np.linalg.norm(A.value() @ X.value() - B.value()) < 1e-12


def test_solve_free_function():
    # 1x1 special case
    check_solve(VariableMatrix([[2.0]]), VariableMatrix([[5.0]]))

    # 2x2 special case
    check_solve(
        VariableMatrix([[1.0, 2.0], [3.0, 4.0]]), VariableMatrix([[5.0], [6.0]])
    )

    # 3x3 special case
    check_solve(
        VariableMatrix([[1.0, 2.0, 3.0], [-4.0, -5.0, 6.0], [7.0, 8.0, 9.0]]),
        VariableMatrix([[10.0], [11.0], [12.0]]),
    )

    # 4x4 special case
    check_solve(
        VariableMatrix(
            [
                [1.0, 2.0, 3.0, -4.0],
                [-5.0, 6.0, 7.0, 8.0],
                [9.0, 10.0, 11.0, 12.0],
                [13.0, 14.0, 15.0, 16.0],
            ]
        ),
        VariableMatrix([[17.0], [18.0], [19.0], [20.0]]),
    )

    # 5x5 general case
    check_solve(
        VariableMatrix(
            [
                [1.0, 2.0, 3.0, -4.0, 5.0],
                [-5.0, 6.0, 7.0, 8.0, 9.0],
                [9.0, 10.0, 11.0, 12.0, 13.0],
                [13.0, 14.0, 15.0, 16.0, 17.0],
                [17.0, 18.0, 19.0, 20.0, 21.0],
            ]
        ),
        VariableMatrix([[21.0], [22.0], [23.0], [24.0], [25.0]]),
    )



import math
from typing import Callable

import numpy as np
import pytest
from sleipnir.autodiff import ExpressionType, VariableMatrix
from sleipnir.optimization import (
    OCP,
    DynamicsType,
    ExitStatus,
    TimestepMethod,
    TranscriptionMethod,
)


def near(expected, actual, tolerance):
    return abs(expected - actual) < tolerance


def flywheel_test(
    test_name: str,
    A: float,
    B: float,
    f: Callable[[VariableMatrix, VariableMatrix], VariableMatrix],
    dynamics_type: DynamicsType,
    transcription_method: TranscriptionMethod,
):
    TOTAL_TIME = 5.0  # s
    dt = 0.005  # s
    N = int(TOTAL_TIME / dt)

    # Flywheel model:
    # States: [velocity]
    # Inputs: [voltage]
    A_discrete = math.exp(A * dt)
    B_discrete = (1.0 - A_discrete) * B

    r = 10.0

    problem = OCP(
        1, 1, dt, N, f, dynamics_type, TimestepMethod.FIXED, transcription_method
    )
    problem.constrain_initial_state(0.0)
    problem.set_upper_input_bound(12)
    problem.set_lower_input_bound(-12)

    # Set up cost
    r_mat = np.full((1, N + 1), r)
    problem.minimize((r_mat - problem.X()) @ (r_mat - problem.X()).T)

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    # Voltage for steady-state velocity:
    #
    # rₖ₊₁ = Arₖ + Buₖ
    # uₖ = B⁺(rₖ₊₁ − Arₖ)
    # uₖ = B⁺(rₖ − Arₖ)
    # uₖ = B⁺(I − A)rₖ
    u_ss = 1.0 / B_discrete * (1.0 - A_discrete) * r

    # Verify initial state
    assert problem.X().value(0, 0) == pytest.approx(0.0, abs=1e-8)

    # Verify solution
    x = 0.0
    u = 0.0
    for k in range(N):
        # Verify state
        assert problem.X().value(0, k) == pytest.approx(x, abs=1e-2)

        # Determine expected input for this timestep
        error = r - x
        if error > 1e-2:
            # Max control input until the reference is reached
            u = 12.0
        else:
            # Maintain speed
            u = u_ss

        # Verify input
        if (
            k > 0
            and k < N - 1
            and near(12.0, problem.U().value(0, k - 1), 1e-2)
            and near(u_ss, problem.U().value(0, k + 1), 1e-2)
        ):
            # If control input is transitioning between 12 and u_ss, ensure it's
            # within (u_ss, 12)
            assert problem.U().value(0, k) >= u_ss
            assert problem.U().value(0, k) <= 12.0
        else:
            if transcription_method == TranscriptionMethod.DIRECT_COLLOCATION:
                # The tolerance is large because the trajectory is represented by a
                # spline, and splines chatter when transitioning quickly between
                # steady-states.
                assert problem.U().value(0, k) == pytest.approx(u, abs=2.0)
            else:
                assert problem.U().value(0, k) == pytest.approx(u, abs=1e-4)

        # Project state forward
        x = A_discrete * x + B_discrete * u

    # Verify final state
    assert problem.X().value(0, N) == pytest.approx(r, abs=1e-6)

    # Log states for offline viewing
    with open(f"{test_name} states.csv", "w") as f:
        f.write("Time (s),Velocity (rad/s)\n")

        for k in range(N + 1):
            f.write(f"{k * dt},{problem.X().value(0, k)}\n")

    # Log inputs for offline viewing
    with open(f"{test_name} inputs.csv", "w") as f:
        f.write("Time (s),Voltage (V)\n")

        for k in range(N + 1):
            if k < N:
                f.write(f"{k * dt},{problem.U().value(0, k)}\n")
            else:
                f.write(f"{k * dt},0.0\n")


def test_flywheel_explicit_ocp():
    A = -1.0
    B = 1.0

    def f_ode(x: VariableMatrix, u: VariableMatrix):
        return A * x + B * u

    flywheel_test(
        "OCP Flywheel Explicit Collocation",
        A,
        B,
        f_ode,
        DynamicsType.EXPLICIT_ODE,
        TranscriptionMethod.DIRECT_COLLOCATION,
    )
    flywheel_test(
        "OCP Flywheel Explicit Transcription",
        A,
        B,
        f_ode,
        DynamicsType.EXPLICIT_ODE,
        TranscriptionMethod.DIRECT_TRANSCRIPTION,
    )
    flywheel_test(
        "OCP Flywheel Explicit Single-Shooting",
        A,
        B,
        f_ode,
        DynamicsType.EXPLICIT_ODE,
        TranscriptionMethod.SINGLE_SHOOTING,
    )


def test_flywheel_discrete_ocp():
    A = -1.0
    B = 1.0
    dt = 0.005  # s

    A_discrete = math.exp(A * dt)
    B_discrete = (1.0 - A_discrete) * B

    def f_discrete(x: VariableMatrix, u: VariableMatrix):
        return A_discrete * x + B_discrete * u

    flywheel_test(
        "OCP Flywheel Discrete Transcription",
        A,
        B,
        f_discrete,
        DynamicsType.DISCRETE,
        TranscriptionMethod.DIRECT_TRANSCRIPTION,
    )
    flywheel_test(
        "OCP Flywheel Discrete Single-Shooting",
        A,
        B,
        f_discrete,
        DynamicsType.DISCRETE,
        TranscriptionMethod.SINGLE_SHOOTING,
    )



import math

import numpy as np
import pytest
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def near(expected, actual, tolerance):
    return abs(expected - actual) < tolerance


def test_flywheel_problem():
    TOTAL_TIME = 5.0
    dt = 0.005
    N = int(TOTAL_TIME / dt)

    # Flywheel model:
    # States: [velocity]
    # Inputs: [voltage]
    A = math.exp(-dt)
    B = 1.0 - math.exp(-dt)

    problem = Problem()
    X = problem.decision_variable(1, N + 1)
    U = problem.decision_variable(1, N)

    # Dynamics constraint
    for k in range(N):
        problem.subject_to(
            X[:, k + 1 : k + 2] == A * X[:, k : k + 1] + B * U[:, k : k + 1]
        )

    # State and input constraints
    problem.subject_to(X[0, 0] == 0.0)
    problem.subject_to(-12 <= U)
    problem.subject_to(U <= 12)

    # Cost function - minimize error
    r = np.array([[10.0]])
    problem.minimize(
        sum((r - X[:, k : k + 1]).T @ (r - X[:, k : k + 1]) for k in range(N + 1))
    )

    assert problem.cost_function_type() == ExpressionType.QUADRATIC
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    # Voltage for steady-state velocity:
    #
    # rₖ₊₁ = Arₖ + Buₖ
    # uₖ = B⁺(rₖ₊₁ − Arₖ)
    # uₖ = B⁺(rₖ − Arₖ)
    # uₖ = B⁺(I − A)rₖ
    u_ss = 1.0 / B * (1.0 - A) * r[0, 0]

    # Verify initial state
    assert X.value(0, 0) == pytest.approx(0.0, abs=1e-8)

    # Verify solution
    x = 0.0
    u = 0.0
    for k in range(N):
        # Verify state
        assert X.value(0, k) == pytest.approx(x, abs=1e-2)

        # Determine expected input for this timestep
        error = r[0, 0] - x
        if error > 1e-2:
            # Max control input until the reference is reached
            u = 12.0
        else:
            # Maintain speed
            u = u_ss

        # Verify input
        if (
            k > 0
            and k < N - 1
            and near(12.0, U.value(0, k - 1), 1e-2)
            and near(u_ss, U.value(0, k + 1), 1e-2)
        ):
            # If control input is transitioning between 12 and u_ss, ensure it's
            # within (u_ss, 12)
            assert U.value(0, k) >= u_ss
            assert U.value(0, k) <= 12.0
        else:
            assert U.value(0, k) == pytest.approx(u, abs=1e-4)

        # Project state forward
        x = A * x + B * u

    # Verify final state
    assert X.value(0, N) == pytest.approx(r[0, 0], abs=1e-7)

    # Log states for offline viewing
    with open("Flywheel states.csv", "w") as f:
        f.write("Time (s),Velocity (rad/s)\n")

        for k in range(N + 1):
            f.write(f"{k * dt},{X.value(0, k)}\n")

    # Log inputs for offline viewing
    with open("Flywheel inputs.csv", "w") as f:
        f.write("Time (s),Voltage (V)\n")

        for k in range(N + 1):
            if k < N:
                f.write(f"{k * dt},{U.value(0, k)}\n")
            else:
                f.write(f"{k * dt},0.0\n")



import platform

import numpy as np
import pytest
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def test_quartic():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(20.0)

    problem.minimize(autodiff.pow(x, 4))

    problem.subject_to(x >= 1)

    assert problem.cost_function_type() == ExpressionType.NONLINEAR
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(1.0, abs=1e-6)


def test_rosenbrock_with_cubic_and_line_constraint():
    # https://en.wikipedia.org/wiki/Test_functions_for_optimization#Test_functions_for_constrained_optimization
    for x0 in np.arange(-1.5, 1.5, 0.1):
        for y0 in np.arange(-0.5, 2.5, 0.1):
            problem = Problem()

            x = problem.decision_variable()
            x.set_value(x0)
            y = problem.decision_variable()
            y.set_value(y0)

            problem.minimize(
                100 * autodiff.pow(y - autodiff.pow(x, 2), 2) + autodiff.pow(1 - x, 2)
            )

            problem.subject_to(y >= autodiff.pow(x - 1, 3) + 1)
            problem.subject_to(y <= -x + 2)

            assert problem.cost_function_type() == ExpressionType.NONLINEAR
            assert problem.equality_constraint_type() == ExpressionType.NONE
            assert problem.inequality_constraint_type() == ExpressionType.NONLINEAR

            assert problem.solve() == ExitStatus.SUCCESS

            # Local minimum at (0.0, 0.0)
            # Global minimum at (1.0, 1.0)
            assert x.value() == pytest.approx(
                0.0, abs=1e-2
            ) or x.value() == pytest.approx(1.0, abs=1e-2)
            assert y.value() == pytest.approx(
                0.0, abs=1e-2
            ) or y.value() == pytest.approx(1.0, abs=1e-2)


def test_rosenbrock_with_disk_constraint():
    # https://en.wikipedia.org/wiki/Test_functions_for_optimization#Test_functions_for_constrained_optimization
    for x0 in np.arange(-1.5, 1.5, 0.1):
        for y0 in np.arange(-1.5, 1.5, 0.1):
            problem = Problem()

            x = problem.decision_variable()
            x.set_value(x0)
            y = problem.decision_variable()
            y.set_value(y0)

            problem.minimize(
                100 * autodiff.pow(y - autodiff.pow(x, 2), 2) + autodiff.pow(1 - x, 2)
            )

            problem.subject_to(autodiff.pow(x, 2) + autodiff.pow(y, 2) <= 2)

            assert problem.cost_function_type() == ExpressionType.NONLINEAR
            assert problem.equality_constraint_type() == ExpressionType.NONE
            assert problem.inequality_constraint_type() == ExpressionType.QUADRATIC

            assert problem.solve() == ExitStatus.SUCCESS

            assert x.value() == pytest.approx(1.0, abs=1e-3)
            assert y.value() == pytest.approx(1.0, abs=1e-3)


def test_minimum_2d_distance_with_linear_constraint():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(20.0)

    y = problem.decision_variable()
    y.set_value(50.0)

    problem.minimize(autodiff.sqrt(x * x + y * y))

    problem.subject_to(y == -x + 5.0)

    assert problem.cost_function_type() == ExpressionType.NONLINEAR
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.NONE

    if platform.system() == "Linux" and platform.machine() == "aarch64":
        # FIXME: Fails on Linux aarch64 with "line search failed"
        assert problem.solve(diagnostics=True) == ExitStatus.LINE_SEARCH_FAILED
        return
    else:
        assert problem.solve(diagnostics=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(2.5, abs=1e-2)
    assert y.value() == pytest.approx(2.5, abs=1e-2)


def test_conflicting_bounds():
    problem = Problem()

    x = problem.decision_variable()
    y = problem.decision_variable()

    problem.minimize(autodiff.hypot(x, y))

    problem.subject_to(autodiff.hypot(x, y) <= 1)
    problem.subject_to(x >= 0.5)
    problem.subject_to(x <= -0.5)

    assert problem.cost_function_type() == ExpressionType.NONLINEAR
    assert problem.equality_constraint_type() == ExpressionType.NONE
    assert problem.inequality_constraint_type() == ExpressionType.NONLINEAR

    assert problem.solve(diagnostics=True) == ExitStatus.GLOBALLY_INFEASIBLE


def test_wachter_and_biegler_line_search_failure():
    # See example 19.2 of [1]
    #
    # [1] Nocedal, J. and Wright, S. "Numerical Optimization", 2nd. ed., Ch. 19.
    #     Springer, 2006.

    problem = Problem()

    x = problem.decision_variable()
    s1 = problem.decision_variable()
    s2 = problem.decision_variable()

    x.set_value(-2)
    s1.set_value(3)
    s2.set_value(1)

    problem.minimize(x)

    problem.subject_to(x**2 - s1 - 1 == 0)
    problem.subject_to(x - s2 - 0.5 == 0)
    problem.subject_to(s1 >= 0)
    problem.subject_to(s2 >= 0)

    assert problem.cost_function_type() == ExpressionType.LINEAR
    assert problem.equality_constraint_type() == ExpressionType.QUADRATIC
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    # FIXME: Fails with "line search failed"
    assert problem.solve(diagnostics=True) == ExitStatus.LINE_SEARCH_FAILED

    # assert x.value() == 1.0
    # assert s1.value() == 0.0
    # assert s2.value() == 0.5



import scipy.sparse

from .._sleipnir.autodiff import *


def spy(mat: scipy.sparse.csc_matrix):
    """
    Plot the sparsity pattern of a sparse matrix.

    Green points represent positive values and red points represent negative
    values.

    Parameter ``mat``:
        The sparse matrix.
    """
    import matplotlib.pyplot as plt
    import matplotlib.ticker as mticker
    from matplotlib.colors import ListedColormap

    xs = []
    ys = []
    vs = []

    cmat = mat.tocoo()
    for row, col, value in zip(cmat.row, cmat.col, cmat.data):
        xs.append(col)
        ys.append(row)
        if value > 0.0:
            vs.append(1.0)
        elif value < 0.0:
            vs.append(-1.0)
        else:
            vs.append(0.0)

    fig = plt.figure()
    ax = fig.add_subplot()

    # Display scatter plot
    cmap = ListedColormap(["red", "green"])
    sc = ax.scatter(xs, ys, s=1, c=vs, marker=".", cmap=cmap, vmin=-1, vmax=1)

    # Display colorbar
    colorbar = fig.colorbar(sc)
    ticklabels = ["Negative", "Positive"]
    colorbar.ax.yaxis.set_major_locator(mticker.FixedLocator(ticklabels))
    colorbar.ax.set_yticks([-1, 1])
    colorbar.ax.set_yticklabels(ticklabels)

    ax.set_title("Sparsity")
    ax.set_xlabel("Cols")
    ax.set_ylabel("Rows")

    ax.set_xlim([0, mat.shape[1]])
    ax.set_ylim([0, mat.shape[0]])
    ax.invert_yaxis()
    ax.set_aspect(1.0)

    plt.show()



import math
import operator
from functools import reduce

import numpy as np
import pytest
import sleipnir.autodiff as autodiff
from sleipnir.autodiff import Gradient, Hessian, Jacobian, Variable, VariableMatrix


def prod(iterable):
    return reduce(operator.mul, iterable, 1)


def test_linear():
    # y = x
    x = VariableMatrix(1)
    x[0].set_value(3)
    y = x[0]

    # dy/dx = 1
    g = Gradient(y, x[0]).value()[0, 0]
    assert g == 1.0

    # d²y/dx² = 0
    H = Hessian(y, x)
    assert H.get().value()[0, 0] == 0.0
    assert H.value()[0, 0] == 0.0


def test_quadratic():
    # y = x²
    x = VariableMatrix(1)
    x[0].set_value(3)
    y = x[0] * x[0]

    # dy/dx = 2x = 6
    g = Gradient(y, x[0]).value()[0, 0]
    assert g == 6.0

    # d²y/dx² = 2
    H = Hessian(y, x)
    assert H.get().value()[0, 0] == 2.0
    assert H.value()[0, 0] == 2.0


def test_cubic():
    # y = x³
    x = VariableMatrix(1)
    x[0].set_value(3)
    y = x[0] * x[0] * x[0]

    # dy/dx = 3x² = 27
    g = Gradient(y, x[0]).value()[0, 0]
    assert g == 27.0

    # d²y/dx² = 6x = 18
    H = Hessian(y, x)
    assert H.get().value()[0, 0] == 18.0
    assert H.value()[0, 0] == 18.0


def test_quartic():
    # y = x⁴
    x = VariableMatrix(1)
    x[0].set_value(3)
    y = x[0] * x[0] * x[0] * x[0]

    # dy/dx = 4x³ = 108
    g = Gradient(y, x[0]).value()[0, 0]
    assert g == 108.0

    # d²y/dx² = 12x² = 108
    H = Hessian(y, x)
    assert H.get().value()[0, 0] == 108.0
    assert H.value()[0, 0] == 108.0


def test_sum():
    y = Variable()
    x = VariableMatrix(5)
    for i in range(5):
        x[i].set_value(i + 1)

    y = sum(x)
    assert y.value() == 15.0

    g = Gradient(y, x)
    assert (g.get().value() == np.full((5, 1), 1.0)).all()
    assert (g.value() == np.full((5, 1), 1.0)).all()

    H = Hessian(y, x)
    assert (H.get().value() == np.zeros((5, 5))).all()
    assert (H.value() == np.zeros((5, 5))).all()


def test_sum_of_products():
    x = VariableMatrix(5)
    for i in range(5):
        x[i].set_value(i + 1)

    # y = ||x||²
    y = x.T @ x
    assert y.value() == sum([x * x for x in range(1, 6)])

    g = Gradient(y, x)
    assert (g.get().value() == 2 * x.value()).all()
    assert (g.value() == 2 * x.value()).all()

    H = Hessian(y, x)

    expected_H = np.diag([2.0] * 5)
    assert (H.get().value() == expected_H).all()
    assert (H.value() == expected_H).all()


def test_product_of_sines():
    x = VariableMatrix(5)
    for i in range(5):
        x[i].set_value(i + 1)

    # y = prod(sin(x))
    y = prod(x.cwise_map(autodiff.sin))
    assert y.value() == pytest.approx(prod(math.sin(x) for x in range(1, 6)), abs=1e-15)

    g = Gradient(y, x)
    for i in range(x.rows()):
        assert g.get().value()[i, 0] == pytest.approx(
            y.value() / math.tan(x[i].value()), abs=1e-15
        )
        assert g.value()[i, 0] == pytest.approx(
            y.value() / math.tan(x[i].value()), abs=1e-15
        )

    H = Hessian(y, x)

    expected_H = np.empty((5, 5))
    for i in range(x.rows()):
        for j in range(x.rows()):
            if i == j:
                expected_H[i, j] = -y.value()
            else:
                expected_H[i, j] = y.value() / (
                    math.tan(x[i].value()) * math.tan(x[j].value())
                )
    np.testing.assert_allclose(H.get().value(), expected_H)
    np.testing.assert_allclose(H.value().todense(), expected_H)


def test_sum_of_squared_residuals():
    x = VariableMatrix(5)
    for i in range(5):
        x[i].set_value(1)

    # y = sum(diff(x).^2)
    y = sum((x[:4, :1] - x[1:5, :1]).cwise_map(lambda x: x**2))
    g = Gradient(y, x).value()

    assert y.value() == 0.0
    assert g[0, 0] == 2 * x[0].value() - 2 * x[1].value()
    assert g[1, 0] == -2 * x[0].value() + 4 * x[1].value() - 2 * x[2].value()
    assert g[2, 0] == -2 * x[1].value() + 4 * x[2].value() - 2 * x[3].value()
    assert g[3, 0] == -2 * x[2].value() + 4 * x[3].value() - 2 * x[4].value()
    assert g[4, 0] == -2 * x[3].value() + 2 * x[4].value()

    H = Hessian(y, x)

    expected_H = np.array(
        [
            [2.0, -2.0, 0.0, 0.0, 0.0],
            [-2.0, 4.0, -2.0, 0.0, 0.0],
            [0.0, -2.0, 4.0, -2.0, 0.0],
            [0.0, 0.0, -2.0, 4.0, -2.0],
            [0.0, 0.0, 0.0, -2.0, 2.0],
        ]
    )
    assert (H.get().value() == expected_H).all()
    assert (H.value().todense() == expected_H).all()


def test_sum_of_squares():
    r = VariableMatrix(4)
    r.set_value(np.array([[25.0], [10.0], [5.0], [0.0]]))

    x = VariableMatrix(4)
    for i in range(4):
        x[i].set_value(0.0)

    J = sum((r[i] - x[i]) * (r[i] - x[i]) for i in range(4))
    H = Hessian(J, x)

    expected_H = np.diag([2.0] * 4)
    assert (H.get().value() == expected_H).all()
    assert (H.value() == expected_H).all()


def test_nested_powers():
    x0 = 3.0

    x = Variable()
    x.set_value(x0)

    y = (x**2) ** 2

    J = Jacobian(y, x).value()
    assert J[0, 0] == pytest.approx(4 * x0 * x0 * x0, abs=1e-12)

    H = Hessian(y, x).value()
    assert H[0, 0] == pytest.approx(12 * x0 * x0, abs=1e-12)


def test_rosenbrock():
    input = VariableMatrix(2)
    x = input[0]
    y = input[1]

    for x0 in np.arange(-2.5, 2.5, 0.1):
        for y0 in np.arange(-2.5, 2.5, 0.1):
            x.set_value(x0)
            y.set_value(y0)
            z = (1 - x) ** 2 + 100 * (y - x**2) ** 2

            H = Hessian(z, input).value()
            assert H[0, 0] == pytest.approx(
                -400 * (y0 - x0 * x0) + 800 * x0 * x0 + 2, abs=1e-11
            )
            assert H[0, 1] == -400 * x0
            assert H[1, 0] == -400 * x0
            assert H[1, 1] == 200


def test_edge_pushing_example_1():
    x = VariableMatrix(2)
    x[0].set_value(3)
    x[1].set_value(4)

    # y = (x₀sin(x₁)) x₀
    y = (x[0] * autodiff.sin(x[1])) * x[0]

    # dy/dx = [2x₀sin(x₁)  x₀²cos(x₁)]
    # dy/dx = [ 6sin(4)     9cos(4)  ]
    J = Jacobian(y, x)
    expected_J = np.array([[6.0 * math.sin(4.0), 9.0 * math.cos(4.0)]])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()

    #           [ 2sin(x₁)    2x₀cos(x₁)]
    # d²y/dx² = [2x₀cos(x₁)  −x₀²sin(x₁)]
    #
    #           [2sin(4)   6cos(4)]
    # d²y/dx² = [6cos(4)  −9sin(4)]
    H = Hessian(y, x)
    expected_H = np.array(
        [
            [2.0 * math.sin(4.0), 6.0 * math.cos(4.0)],
            [6.0 * math.cos(4.0), -9.0 * math.sin(4.0)],
        ]
    )
    assert (H.get().value() == expected_H).all()
    assert (H.value() == expected_H).all()


def test_variable_reuse():
    y = Variable()
    x = VariableMatrix(1)

    # y = x³
    x[0].set_value(1)
    y = x[0] * x[0] * x[0]

    hessian = Hessian(y, x)

    # d²y/dx² = 6x
    # H = 6
    H = hessian.value()

    assert H.shape == (1, 1)
    assert H[0, 0] == 6.0

    x[0].set_value(2)
    # d²y/dx² = 6x
    # H = 12
    H = hessian.value()

    assert H.shape == (1, 1)
    assert H[0, 0] == 12.0



import numpy as np
from sleipnir.autodiff import Variable, VariableMatrix
from sleipnir.optimization import EqualityConstraints, InequalityConstraints


def test_equality_constraint_boolean_comparisons():
    args = [(1.0, 1.0), (1.0, 2.0), (2.0, 1.0)]

    # double-Variable
    for lhs, rhs in args:
        assert bool(float(lhs) == Variable(rhs)) == (lhs == rhs)

    # double-VariableMatrix
    for lhs, rhs in args:
        assert bool(float(lhs) == VariableMatrix([[rhs]])) == (lhs == rhs)

    # Variable-double
    for lhs, rhs in args:
        assert bool(Variable(lhs) == float(rhs)) == (lhs == rhs)

    # Variable-Variable
    for lhs, rhs in args:
        assert bool(Variable(lhs) == Variable(rhs)) == (lhs == rhs)

    # Variable-VariableMatrix
    for lhs, rhs in args:
        assert bool(Variable(lhs) == VariableMatrix([[rhs]])) == (lhs == rhs)

    # VariableMatrix-double
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) == float(rhs)) == (lhs == rhs)

    # VariableMatrix-Variable
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) == Variable(rhs)) == (lhs == rhs)

    # VariableMatrix-VariableMatrix
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) == VariableMatrix([[rhs]])) == (lhs == rhs)

    # np.array-VariableMatrix
    for lhs, rhs in args:
        assert bool(np.array([[lhs]]) == VariableMatrix([[rhs]])) == (lhs == rhs)

    # np.array-VariableBlock
    for lhs, rhs in args:
        assert bool(np.array([[lhs]]) == VariableMatrix([[rhs]])[:, :]) == (lhs == rhs)

    # VariableMatrix-np.array
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) == np.array([[rhs]])) == (lhs == rhs)

    # VariableBlock-np.array
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]])[:, :] == np.array([[rhs]])) == (lhs == rhs)


# For the purposes of optimization, a < constraint is treated the same as a <=
# constraint
def test_inequality_constraint_boolean_comparisons():
    args = [(1.0, 1.0), (1.0, 2.0), (2.0, 1.0)]

    # double-Variable
    for lhs, rhs in args:
        assert bool(float(lhs) < Variable(rhs)) == (lhs <= rhs)
        assert bool(float(lhs) <= Variable(rhs)) == (lhs <= rhs)
        assert bool(float(lhs) > Variable(rhs)) == (lhs >= rhs)
        assert bool(float(lhs) >= Variable(rhs)) == (lhs >= rhs)

    # double-VariableMatrix
    for lhs, rhs in args:
        assert bool(float(lhs) < VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(float(lhs) <= VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(float(lhs) > VariableMatrix([[rhs]])) == (lhs >= rhs)
        assert bool(float(lhs) >= VariableMatrix([[rhs]])) == (lhs >= rhs)

    # Variable-Variable
    for lhs, rhs in args:
        assert bool(Variable(lhs) < Variable(rhs)) == (lhs <= rhs)
        assert bool(Variable(lhs) <= Variable(rhs)) == (lhs <= rhs)
        assert bool(Variable(lhs) > Variable(rhs)) == (lhs >= rhs)
        assert bool(Variable(lhs) >= Variable(rhs)) == (lhs >= rhs)

    # Variable-VariableMatrix
    for lhs, rhs in args:
        assert bool(Variable(lhs) < VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(Variable(lhs) <= VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(Variable(lhs) > VariableMatrix([[rhs]])) == (lhs >= rhs)
        assert bool(Variable(lhs) >= VariableMatrix([[rhs]])) == (lhs >= rhs)

    # VariableMatrix-double
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) < float(rhs)) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) <= float(rhs)) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) > float(rhs)) == (lhs >= rhs)
        assert bool(VariableMatrix([[lhs]]) >= float(rhs)) == (lhs >= rhs)

    # VariableMatrix-Variable
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) < Variable(rhs)) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) <= Variable(rhs)) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) > Variable(rhs)) == (lhs >= rhs)
        assert bool(VariableMatrix([[lhs]]) >= Variable(rhs)) == (lhs >= rhs)

    # VariableMatrix-VariableMatrix
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) < VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) <= VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) > VariableMatrix([[rhs]])) == (lhs >= rhs)
        assert bool(VariableMatrix([[lhs]]) >= VariableMatrix([[rhs]])) == (lhs >= rhs)

    # np.array-VariableMatrix
    for lhs, rhs in args:
        assert bool(np.array([[lhs]]) < VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(np.array([[lhs]]) <= VariableMatrix([[rhs]])) == (lhs <= rhs)
        assert bool(np.array([[lhs]]) > VariableMatrix([[rhs]])) == (lhs >= rhs)
        assert bool(np.array([[lhs]]) >= VariableMatrix([[rhs]])) == (lhs >= rhs)

    # np.array-VariableBlock
    for lhs, rhs in args:
        assert bool(np.array([[lhs]]) < VariableMatrix([[rhs]])[:, :]) == (lhs <= rhs)
        assert bool(np.array([[lhs]]) <= VariableMatrix([[rhs]])[:, :]) == (lhs <= rhs)
        assert bool(np.array([[lhs]]) > VariableMatrix([[rhs]])[:, :]) == (lhs >= rhs)
        assert bool(np.array([[lhs]]) >= VariableMatrix([[rhs]])[:, :]) == (lhs >= rhs)

    # VariableMatrix-np.array
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]]) < np.array([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) <= np.array([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]]) > np.array([[rhs]])) == (lhs >= rhs)
        assert bool(VariableMatrix([[lhs]]) >= np.array([[rhs]])) == (lhs >= rhs)

    # VariableBlock-np.array
    for lhs, rhs in args:
        assert bool(VariableMatrix([[lhs]])[:, :] < np.array([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]])[:, :] <= np.array([[rhs]])) == (lhs <= rhs)
        assert bool(VariableMatrix([[lhs]])[:, :] > np.array([[rhs]])) == (lhs >= rhs)
        assert bool(VariableMatrix([[lhs]])[:, :] >= np.array([[rhs]])) == (lhs >= rhs)


def test_equality_constraint_concatenation():
    eq1 = Variable(1.0) == Variable(1.0)
    eq2 = Variable(1.0) == Variable(2.0)
    eqs = EqualityConstraints([eq1, eq2])

    assert bool(eq1)
    assert not bool(eq2)
    assert not bool(eqs)


def test_inequality_constraint_concatenation():
    ineq1 = Variable(2.0) < Variable(1.0)
    ineq2 = Variable(1.0) < Variable(2.0)
    ineqs = InequalityConstraints([ineq1, ineq2])

    assert not bool(ineq1)
    assert bool(ineq2)
    assert not bool(ineqs)



import numpy as np
import pytest
from differential_drive_util import (
    differential_drive_dynamics_double,
    differential_drive_dynamics_variable,
)
from rk4 import rk4
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import (
    OCP,
    DynamicsType,
    ExitStatus,
    TimestepMethod,
    TranscriptionMethod,
)


def test_differential_drive_ocp():
    N = 50

    min_timestep = 0.05  # s
    x_initial = np.zeros((5, 1))
    x_final = np.array([[1.0], [1.0], [0.0], [0.0], [0.0]])
    u_min = np.array([[-12.0], [-12.0]])
    u_max = np.array([[12.0], [12.0]])

    problem = OCP(
        5,
        2,
        min_timestep,
        N,
        differential_drive_dynamics_variable,
        DynamicsType.EXPLICIT_ODE,
        TimestepMethod.VARIABLE_SINGLE,
        TranscriptionMethod.DIRECT_TRANSCRIPTION,
    )

    # Seed the min time formulation with lerp between waypoints
    for i in range(N + 1):
        problem.X()[0, i].set_value(i / (N + 1))
        problem.X()[1, i].set_value(i / (N + 1))

    problem.constrain_initial_state(x_initial)
    problem.constrain_final_state(x_final)

    problem.set_lower_input_bound(u_min)
    problem.set_upper_input_bound(u_max)

    problem.set_min_timestep(min_timestep)
    problem.set_max_timestep(3.0)

    # Set up cost
    problem.minimize(problem.dt() @ np.ones((N + 1, 1)))

    assert problem.cost_function_type() == ExpressionType.LINEAR
    assert problem.equality_constraint_type() == ExpressionType.NONLINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(max_iterations=1000, diagnostics=True) == ExitStatus.SUCCESS

    X = problem.X()
    U = problem.U()

    # Verify initial state
    assert X.value(0, 0) == pytest.approx(x_initial[0, 0], abs=1e-8)
    assert X.value(1, 0) == pytest.approx(x_initial[1, 0], abs=1e-8)
    assert X.value(2, 0) == pytest.approx(x_initial[2, 0], abs=1e-8)
    assert X.value(3, 0) == pytest.approx(x_initial[3, 0], abs=1e-8)
    assert X.value(4, 0) == pytest.approx(x_initial[4, 0], abs=1e-8)

    # FIXME: Replay diverges
    return

    # Verify solution
    x = np.zeros((5, 1))
    u = np.zeros((2, 1))
    for k in range(N):
        u = U[:, k : k + 1].value()

        # Input constraints
        assert U[0, k].value() >= -u_max[0]
        assert U[0, k].value() <= u_max[0]
        assert U[1, k].value() >= -u_max[1]
        assert U[1, k].value() <= u_max[1]

        # Verify state
        assert X.value(0, k) == pytest.approx(x[0, 0], abs=1e-8)
        assert X.value(1, k) == pytest.approx(x[1, 0], abs=1e-8)
        assert X.value(2, k) == pytest.approx(x[2, 0], abs=1e-8)
        assert X.value(3, k) == pytest.approx(x[3, 0], abs=1e-8)
        assert X.value(4, k) == pytest.approx(x[4, 0], abs=1e-8)

        # Project state forward
        x = rk4(differential_drive_dynamics_double, x, u, problem.dt().value(0, k))

    # Verify final state
    assert X.value(0, N) == pytest.approx(x_final[0, 0], abs=1e-8)
    assert X.value(1, N) == pytest.approx(x_final[1, 0], abs=1e-8)
    assert X.value(2, N) == pytest.approx(x_final[2, 0], abs=1e-8)
    assert X.value(3, N) == pytest.approx(x_final[3, 0], abs=1e-8)
    assert X.value(4, N) == pytest.approx(x_final[4, 0], abs=1e-8)

    # Log states for offline viewing
    with open("Differential drive states.csv", "w") as f:
        f.write(
            "Time (s),X position (m),Y position (m),Heading (rad),Left velocity (m/s),Right velocity (m/s)\n"
        )

        time = 0.0
        for k in range(N + 1):
            f.write(
                f"{time},{X.value(0, k)},{X.value(1, k)},{X.value(2, k)},{X.value(3, k)},{X.value(4, k)}\n"
            )

            time += problem.dt().value(0, k)

    # Log inputs for offline viewing
    with open("Differential drive inputs.csv", "w") as f:
        f.write("Time (s),Left voltage (V),Right voltage (V)\n")

        time = 0.0
        for k in range(N + 1):
            if k < N:
                f.write(f"{time},{U.value(0, k)},{U.value(1, k)}\n")
            else:
                f.write(f"{time},0.0,0.0\n")

            time += problem.dt().value(0, k)



import struct
from typing import BinaryIO

import pytest
from sleipnir.autodiff import ExpressionType
from sleipnir.optimization import ExitStatus, Problem


def read_char(spy: BinaryIO) -> bytes:
    return struct.unpack("c", spy.read(1))[0]


def read_i32(spy: BinaryIO) -> int:
    return struct.unpack("<i", spy.read(4))[0]


def read_str(spy: BinaryIO) -> str:
    size = read_i32(spy)
    return struct.unpack(f"{size}s", spy.read(size))[0].decode("utf-8")


class Coord:
    def __init__(self, row: int, col: int, sign: bytes):
        self.row = row
        self.col = col
        self.sign = sign

    def __eq__(self, other) -> bool:
        return (
            self.row == other.row and self.col == other.col and self.sign == other.sign
        )


def read_coord(spy: BinaryIO) -> Coord:
    return Coord(read_i32(spy), read_i32(spy), read_char(spy))


iterations = 0


def test_problem_spy():
    problem = Problem()

    x = problem.decision_variable()
    x.set_value(20.0)
    y = problem.decision_variable()
    y.set_value(20.0)

    problem.minimize(x**4 + y**4)

    problem.subject_to(x >= 1.0)
    problem.subject_to(x <= 10.0)
    problem.subject_to(y == 2.0)

    global iterations
    iterations = 0

    def callback(info):
        global iterations
        iterations += 1
        return False

    problem.add_callback(callback)

    assert problem.cost_function_type() == ExpressionType.NONLINEAR
    assert problem.equality_constraint_type() == ExpressionType.LINEAR
    assert problem.inequality_constraint_type() == ExpressionType.LINEAR

    assert problem.solve(diagnostics=True, spy=True) == ExitStatus.SUCCESS

    assert x.value() == pytest.approx(1.0, abs=1e-8)
    assert y.value() == pytest.approx(2.0, abs=1e-8)

    # Check H.spy
    with open("H.spy", mode="rb") as spy:
        assert read_str(spy) == "Hessian"  # Title
        assert read_str(spy) == "Decision variables"  # Row label
        assert read_str(spy) == "Decision variables"  # Col label
        assert read_i32(spy) == 2  # Rows
        assert read_i32(spy) == 2  # Cols

        # Coords
        for _ in range(iterations):
            assert read_i32(spy) == 2  # Num coords
            assert read_coord(spy) == Coord(0, 0, b"+")
            assert read_coord(spy) == Coord(1, 1, b"+")

    # Check A_e.spy
    with open("A_e.spy", mode="rb") as spy:
        assert read_str(spy) == "Equality constraint Jacobian"  # Title
        assert read_str(spy) == "Constraints"  # Row label
        assert read_str(spy) == "Decision variables"  # Col label
        assert read_i32(spy) == 1  # Rows
        assert read_i32(spy) == 2  # Cols

        # Coords
        for _ in range(iterations):
            assert read_i32(spy) == 1  # Num coords
            assert read_coord(spy) == Coord(0, 1, b"+")

    # Check A_i.spy
    with open("A_i.spy", mode="rb") as spy:
        assert read_str(spy) == "Inequality constraint Jacobian"  # Title
        assert read_str(spy) == "Constraints"  # Row label
        assert read_str(spy) == "Decision variables"  # Col label
        assert read_i32(spy) == 2  # Rows
        assert read_i32(spy) == 2  # Cols

        # Coords
        for _ in range(iterations):
            assert read_i32(spy) == 2  # Num coords
            assert read_coord(spy) == Coord(0, 0, b"+")
            assert read_coord(spy) == Coord(1, 0, b"-")



import numpy as np
from sleipnir.autodiff import Jacobian, VariableMatrix


def test_y_eq_x():
    x = VariableMatrix(3)
    for i in range(3):
        x[i].set_value(i + 1)

    # y = x
    #
    #         [1  0  0]
    # dy/dx = [0  1  0]
    #         [0  0  1]
    y = x
    J = Jacobian(y, x)

    expected_J = np.diag([1.0, 1.0, 1.0])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()


def test_y_eq_3x():
    x = VariableMatrix(3)
    for i in range(3):
        x[i].set_value(i + 1)

    # y = 3x
    #
    #         [3  0  0]
    # dy/dx = [0  3  0]
    #         [0  0  3]
    y = 3 * x
    J = Jacobian(y, x)

    expected_J = np.diag([3.0, 3.0, 3.0])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()


def test_products():
    x = VariableMatrix(3)
    for i in range(3):
        x[i].set_value(i + 1)

    #     [x₁x₂]
    # y = [x₂x₃]
    #     [x₁x₃]
    #
    #         [x₂  x₁  0 ]
    # dy/dx = [0   x₃  x₂]
    #         [x₃  0   x₁]
    #
    #         [2  1  0]
    # dy/dx = [0  3  2]
    #         [3  0  1]
    y = VariableMatrix(3)
    y[0] = x[0] * x[1]
    y[1] = x[1] * x[2]
    y[2] = x[0] * x[2]
    J = Jacobian(y, x)

    expected_J = np.array([[2.0, 1.0, 0.0], [0.0, 3.0, 2.0], [3.0, 0.0, 1.0]])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()


def test_nested_products():
    x = VariableMatrix(1)
    x[0].set_value(3)
    assert x.value(0) == 3.0

    #     [ 5x]   [15]
    # y = [ 7x] = [21]
    #     [11x]   [33]
    y = VariableMatrix(3)
    y[0] = 5 * x[0]
    y[1] = 7 * x[0]
    y[2] = 11 * x[0]
    assert y.value(0) == 15.0
    assert y.value(1) == 21.0
    assert y.value(2) == 33.0

    #     [y₁y₂]   [15⋅21]   [315]
    # z = [y₂y₃] = [21⋅33] = [693]
    #     [y₁y₃]   [15⋅33]   [495]
    z = VariableMatrix(3)
    z[0] = y[0] * y[1]
    z[1] = y[1] * y[2]
    z[2] = y[0] * y[2]
    assert z.value(0) == 315.0
    assert z.value(1) == 693.0
    assert z.value(2) == 495.0

    #     [ 5x]
    # y = [ 7x]
    #     [11x]
    #
    #         [ 5]
    # dy/dx = [ 7]
    #         [11]
    J = Jacobian(y, x)
    assert J.get().value()[0, 0] == 5.0
    assert J.get().value()[1, 0] == 7.0
    assert J.get().value()[2, 0] == 11.0
    assert J.value()[0, 0] == 5.0
    assert J.value()[1, 0] == 7.0
    assert J.value()[2, 0] == 11.0

    #     [y₁y₂]
    # z = [y₂y₃]
    #     [y₁y₃]
    #
    #         [y₂  y₁  0 ]   [21  15   0]
    # dz/dy = [0   y₃  y₂] = [ 0  33  21]
    #         [y₃  0   y₁]   [33   0  15]
    J = Jacobian(z, y)
    expected_J = np.array([[21.0, 15.0, 0.0], [0.0, 33.0, 21.0], [33.0, 0.0, 15.0]])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()

    #     [y₁y₂]   [5x⋅ 7x]   [35x²]
    # z = [y₂y₃] = [7x⋅11x] = [77x²]
    #     [y₁y₃]   [5x⋅11x]   [55x²]
    #
    #         [ 70x]   [210]
    # dz/dx = [154x] = [462]
    #         [110x] = [330]
    J = Jacobian(z, x)
    expected_J = np.array([[210.0], [462.0], [330.0]])
    assert (J.get().value() == expected_J).all()
    assert (J.value() == expected_J).all()


def test_non_square():
    x = VariableMatrix(3)
    for i in range(3):
        x[i].set_value(i + 1)

    # y = [x₁ + 3x₂ − 5x₃]
    #
    # dy/dx = [1  3  −5]
    y = VariableMatrix(1)
    y[0] = x[0] + 3 * x[1] - 5 * x[2]
    J = Jacobian(y, x)

    expected_J = np.array([[1.0, 3.0, -5.0]])

    J_get_value = J.get().value()
    assert J_get_value.shape == (1, 3)
    assert (J_get_value == expected_J).all()

    J_value = J.get().value()
    assert J_value.shape == (1, 3)
    assert (J_value == expected_J).all()


def test_variable_reuse():
    x = VariableMatrix(2)
    for i in range(2):
        x[i].set_value(i + 1)

    # y = [x₁x₂]
    y = VariableMatrix(1)
    y[0] = x[0] * x[1]

    jacobian = Jacobian(y, x)

    # dy/dx = [x₂  x₁]
    # dy/dx = [2  1]
    J = jacobian.value()

    assert J.shape == (1, 2)
    assert J[0, 0] == 2.0
    assert J[0, 1] == 1.0

    x[0].set_value(2)
    x[1].set_value(1)
    # dy/dx = [x₂  x₁]
    # dy/dx = [1  2]
    J = jacobian.value()

    assert J.shape == (1, 2)
    assert J[0, 0] == 1.0
    assert J[0, 1] == 2.0
